# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d

from pathlib import Path
import pytest


def test_dataset_base():
    default_data_root = Path.home() / "open3d_data"

    ds = o3d.data.Dataset("some_prefix")
    assert Path(ds.data_root) == default_data_root

    ds_custom = o3d.data.Dataset("some_prefix", "/my/custom/data_root")
    assert Path(ds_custom.data_root) == Path("/my/custom/data_root")
    assert ds_custom.prefix == "some_prefix"
    assert ds_custom.download_dir == "/my/custom/data_root/download/some_prefix"
    assert ds_custom.extract_dir == "/my/custom/data_root/extract/some_prefix"


def get_test_data_dirs(prefix):
    gt_data_root = Path.home() / "open3d_data"
    gt_download_dir = gt_data_root / "download" / prefix
    gt_extract_dir = gt_data_root / "extract" / prefix
    return gt_data_root, gt_download_dir, gt_extract_dir


def test_simple_dataset_base():
    gt_prefix = "BunnyMesh"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data_descriptor = o3d.data.DataDescriptor(
        url=o3d.data.open3d_downloads_prefix + "20220201-data/BunnyMesh.ply",
        md5="568f871d1a221ba6627569f1e6f9a3f2")
    single_download_dataset = o3d.data.DownloadDataset(
        prefix=gt_prefix,
        data_descriptor=data_descriptor,
    )

    assert Path(gt_extract_dir / "BunnyMesh.ply").is_file()
    assert Path(gt_download_dir / "BunnyMesh.ply").is_file()

    assert single_download_dataset.prefix == gt_prefix
    assert Path(single_download_dataset.data_root) == gt_data_root
    assert Path(single_download_dataset.download_dir) == gt_download_dir
    assert Path(single_download_dataset.extract_dir) == gt_extract_dir


def test_demo_icp_pointclouds():
    gt_prefix = "DemoICPPointClouds"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    demo_icp = o3d.data.DemoICPPointClouds()
    assert Path(gt_download_dir).is_dir()

    gt_paths = [
        gt_extract_dir / "cloud_bin_0.pcd", gt_extract_dir / "cloud_bin_1.pcd",
        gt_extract_dir / "cloud_bin_2.pcd"
    ]

    assert len(demo_icp.paths) == len(gt_paths)
    for gt_path, demo_icp_path in zip(gt_paths, demo_icp.paths):
        assert Path(gt_path) == Path(demo_icp_path)
        assert Path(gt_path).is_file()

    assert Path(demo_icp.transformation_log_path) == Path(gt_extract_dir /
                                                          "init.log")
    assert Path(demo_icp.transformation_log_path).is_file()

    assert demo_icp.prefix == gt_prefix
    assert Path(demo_icp.data_root) == gt_data_root
    assert Path(demo_icp.download_dir) == gt_download_dir
    assert Path(demo_icp.extract_dir) == gt_extract_dir


def test_demo_colored_icp_pointclouds():
    gt_prefix = "DemoColoredICPPointClouds"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    demo_colored_icp = o3d.data.DemoColoredICPPointClouds()
    assert Path(gt_download_dir).is_dir()

    gt_paths = [
        gt_extract_dir / "frag_115.ply", gt_extract_dir / "frag_116.ply"
    ]

    assert len(demo_colored_icp.paths) == len(gt_paths)
    for gt_path, demo_colored_icp_path in zip(gt_paths, demo_colored_icp.paths):
        assert Path(gt_path) == Path(demo_colored_icp_path)
        assert Path(gt_path).is_file()

    assert demo_colored_icp.prefix == gt_prefix
    assert Path(demo_colored_icp.data_root) == gt_data_root
    assert Path(demo_colored_icp.download_dir) == gt_download_dir
    assert Path(demo_colored_icp.extract_dir) == gt_extract_dir


def test_demo_crop_pointcloud():
    gt_prefix = "DemoCropPointCloud"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    demo_crop_pcd = o3d.data.DemoCropPointCloud()
    assert Path(gt_download_dir).is_dir()

    assert Path(demo_crop_pcd.point_cloud_path) == gt_extract_dir / \
        "fragment.ply"
    assert Path(demo_crop_pcd.point_cloud_path).is_file()
    assert Path(demo_crop_pcd.cropped_json_path) == gt_extract_dir / \
        "cropped.json"
    assert Path(demo_crop_pcd.point_cloud_path).is_file()

    assert demo_crop_pcd.prefix == gt_prefix
    assert Path(demo_crop_pcd.data_root) == gt_data_root
    assert Path(demo_crop_pcd.download_dir) == gt_download_dir
    assert Path(demo_crop_pcd.extract_dir) == gt_extract_dir


def test_demo_feature_matching_point_clouds():
    gt_prefix = "DemoFeatureMatchingPointClouds"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    demo_feature_matching = o3d.data.DemoFeatureMatchingPointClouds()
    assert Path(gt_download_dir).is_dir()

    gt_point_cloud_paths = [
        gt_extract_dir / "cloud_bin_0.pcd", gt_extract_dir / "cloud_bin_1.pcd"
    ]
    assert len(
        demo_feature_matching.point_cloud_paths) == len(gt_point_cloud_paths)
    for gt_path, demo_feature_matching_point_cloud_path in zip(
            gt_point_cloud_paths, demo_feature_matching.point_cloud_paths):
        assert Path(demo_feature_matching_point_cloud_path) == gt_path
        assert Path(gt_path).is_file()

    gt_fpfh_feature_paths = [
        gt_extract_dir / "cloud_bin_0.fpfh.bin",
        gt_extract_dir / "cloud_bin_1.fpfh.bin"
    ]
    assert len(
        demo_feature_matching.fpfh_feature_paths) == len(gt_fpfh_feature_paths)
    for gt_path, demo_feature_matching_fpfh_feature_path in zip(
            gt_fpfh_feature_paths, demo_feature_matching.fpfh_feature_paths):
        assert Path(demo_feature_matching_fpfh_feature_path) == gt_path
        assert Path(gt_path).is_file()

    gt_l32d_feature_paths = [
        gt_extract_dir / "cloud_bin_0.d32.bin",
        gt_extract_dir / "cloud_bin_1.d32.bin"
    ]
    assert len(
        demo_feature_matching.l32d_feature_paths) == len(gt_l32d_feature_paths)
    for gt_path, demo_feature_matching_l32d_feature_path in zip(
            gt_l32d_feature_paths, demo_feature_matching.l32d_feature_paths):
        assert Path(demo_feature_matching_l32d_feature_path) == gt_path
        assert Path(gt_path).is_file()

    assert demo_feature_matching.prefix == gt_prefix
    assert Path(demo_feature_matching.data_root) == gt_data_root
    assert Path(demo_feature_matching.download_dir) == gt_download_dir
    assert Path(demo_feature_matching.extract_dir) == gt_extract_dir


def test_demo_pose_graph_optimization():
    gt_prefix = "DemoPoseGraphOptimization"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    demo_pose_optimization = o3d.data.DemoPoseGraphOptimization()
    assert Path(gt_download_dir).is_dir()

    assert Path(demo_pose_optimization.pose_graph_fragment_path) == gt_extract_dir / \
        "pose_graph_example_fragment.json"
    assert Path(demo_pose_optimization.pose_graph_fragment_path).is_file()
    assert Path(demo_pose_optimization.pose_graph_global_path) == gt_extract_dir / \
        "pose_graph_example_global.json"
    assert Path(demo_pose_optimization.pose_graph_global_path).is_file()

    assert demo_pose_optimization.prefix == gt_prefix
    assert Path(demo_pose_optimization.data_root) == gt_data_root
    assert Path(demo_pose_optimization.download_dir) == gt_download_dir
    assert Path(demo_pose_optimization.extract_dir) == gt_extract_dir


def test_pcd_point_cloud():
    gt_prefix = "PCDPointCloud"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    pcd_pointcloud = o3d.data.PCDPointCloud()
    assert Path(gt_download_dir).is_dir()

    assert Path(pcd_pointcloud.path) == gt_extract_dir / "fragment.pcd"
    assert Path(pcd_pointcloud.path).is_file()

    assert pcd_pointcloud.prefix == gt_prefix
    assert Path(pcd_pointcloud.data_root) == gt_data_root
    assert Path(pcd_pointcloud.download_dir) == gt_download_dir
    assert Path(pcd_pointcloud.extract_dir) == gt_extract_dir


def test_ply_point_cloud():
    gt_prefix = "PLYPointCloud"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    ply_pointcloud = o3d.data.PLYPointCloud()
    assert Path(gt_download_dir).is_dir()

    assert Path(ply_pointcloud.path) == gt_extract_dir / "fragment.ply"
    assert Path(ply_pointcloud.path).is_file()

    assert ply_pointcloud.prefix == gt_prefix
    assert Path(ply_pointcloud.data_root) == gt_data_root
    assert Path(ply_pointcloud.download_dir) == gt_download_dir
    assert Path(ply_pointcloud.extract_dir) == gt_extract_dir


def test_sample_nyu_rgbd_image():
    gt_prefix = "SampleNYURGBDImage"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    rgbd_image_nyu = o3d.data.SampleNYURGBDImage()
    assert Path(gt_download_dir).is_dir()

    assert Path(rgbd_image_nyu.color_path) == gt_extract_dir / "NYU_color.ppm"
    assert Path(rgbd_image_nyu.color_path).is_file()

    assert Path(rgbd_image_nyu.depth_path) == gt_extract_dir / "NYU_depth.pgm"
    assert Path(rgbd_image_nyu.depth_path).is_file()

    assert rgbd_image_nyu.prefix == gt_prefix
    assert Path(rgbd_image_nyu.data_root) == gt_data_root
    assert Path(rgbd_image_nyu.download_dir) == gt_download_dir
    assert Path(rgbd_image_nyu.extract_dir) == gt_extract_dir


def test_sample_sun_rgbd_image():
    gt_prefix = "SampleSUNRGBDImage"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    rgbd_image_sun = o3d.data.SampleSUNRGBDImage()
    assert Path(gt_download_dir).is_dir()

    assert Path(rgbd_image_sun.color_path) == gt_extract_dir / "SUN_color.jpg"
    assert Path(rgbd_image_sun.color_path).is_file()

    assert Path(rgbd_image_sun.depth_path) == gt_extract_dir / "SUN_depth.png"
    assert Path(rgbd_image_sun.depth_path).is_file()

    assert rgbd_image_sun.prefix == gt_prefix
    assert Path(rgbd_image_sun.data_root) == gt_data_root
    assert Path(rgbd_image_sun.download_dir) == gt_download_dir
    assert Path(rgbd_image_sun.extract_dir) == gt_extract_dir


def test_sample_tum_rgbd_image():
    gt_prefix = "SampleTUMRGBDImage"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    rgbd_image_tum = o3d.data.SampleTUMRGBDImage()
    assert Path(gt_download_dir).is_dir()

    assert Path(rgbd_image_tum.color_path) == gt_extract_dir / "TUM_color.png"
    assert Path(rgbd_image_tum.color_path).is_file()

    assert Path(rgbd_image_tum.depth_path) == gt_extract_dir / "TUM_depth.png"
    assert Path(rgbd_image_tum.depth_path).is_file()

    assert rgbd_image_tum.prefix == gt_prefix
    assert Path(rgbd_image_tum.data_root) == gt_data_root
    assert Path(rgbd_image_tum.download_dir) == gt_download_dir
    assert Path(rgbd_image_tum.extract_dir) == gt_extract_dir


def test_sample_redwood_rgbd_images():
    gt_prefix = "SampleRedwoodRGBDImages"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    rgbd_dataset_redwood = o3d.data.SampleRedwoodRGBDImages()
    assert Path(gt_download_dir).is_dir()

    gt_color_paths = [
        gt_extract_dir / "color" / "00000.jpg",
        gt_extract_dir / "color" / "00001.jpg",
        gt_extract_dir / "color" / "00002.jpg",
        gt_extract_dir / "color" / "00003.jpg",
        gt_extract_dir / "color" / "00004.jpg"
    ]
    assert len(rgbd_dataset_redwood.color_paths) == len(gt_color_paths)
    for gt_path, rgbd_dataset_redwood_color_path in zip(
            gt_color_paths, rgbd_dataset_redwood.color_paths):
        assert Path(rgbd_dataset_redwood_color_path) == gt_path
        assert Path(gt_path).is_file()

    gt_depth_paths = [
        gt_extract_dir / "depth" / "00000.png",
        gt_extract_dir / "depth" / "00001.png",
        gt_extract_dir / "depth" / "00002.png",
        gt_extract_dir / "depth" / "00003.png",
        gt_extract_dir / "depth" / "00004.png"
    ]
    assert len(rgbd_dataset_redwood.depth_paths) == len(gt_depth_paths)
    for gt_path, rgbd_dataset_redwood_depth_path in zip(
            gt_depth_paths, rgbd_dataset_redwood.depth_paths):
        assert Path(rgbd_dataset_redwood_depth_path) == gt_path
        assert Path(gt_path).is_file()

    assert Path(rgbd_dataset_redwood.trajectory_log_path
               ) == gt_extract_dir / "trajectory.log"
    assert Path(rgbd_dataset_redwood.odometry_log_path
               ) == gt_extract_dir / "odometry.log"
    assert Path(
        rgbd_dataset_redwood.rgbd_match_path) == gt_extract_dir / "rgbd.match"
    assert Path(rgbd_dataset_redwood.reconstruction_path
               ) == gt_extract_dir / "example_tsdf_pcd.ply"

    assert rgbd_dataset_redwood.prefix == gt_prefix
    assert Path(rgbd_dataset_redwood.data_root) == gt_data_root
    assert Path(rgbd_dataset_redwood.download_dir) == gt_download_dir
    assert Path(rgbd_dataset_redwood.extract_dir) == gt_extract_dir


def test_sample_fountain_rgbd_images():
    gt_prefix = "SampleFountainRGBDImages"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    fountain_dataset = o3d.data.SampleFountainRGBDImages()
    assert Path(gt_download_dir).is_dir()

    gt_color_paths = [
        gt_extract_dir / "image" / "0000010-000001228920.jpg",
        gt_extract_dir / "image" / "0000031-000004096400.jpg",
        gt_extract_dir / "image" / "0000044-000005871507.jpg",
        gt_extract_dir / "image" / "0000064-000008602440.jpg",
        gt_extract_dir / "image" / "0000110-000014883587.jpg",
        gt_extract_dir / "image" / "0000156-000021164733.jpg",
        gt_extract_dir / "image" / "0000200-000027172787.jpg",
        gt_extract_dir / "image" / "0000215-000029220987.jpg",
        gt_extract_dir / "image" / "0000255-000034682853.jpg",
        gt_extract_dir / "image" / "0000299-000040690907.jpg",
        gt_extract_dir / "image" / "0000331-000045060400.jpg",
        gt_extract_dir / "image" / "0000368-000050112627.jpg",
        gt_extract_dir / "image" / "0000412-000056120680.jpg",
        gt_extract_dir / "image" / "0000429-000058441973.jpg",
        gt_extract_dir / "image" / "0000474-000064586573.jpg",
        gt_extract_dir / "image" / "0000487-000066361680.jpg",
        gt_extract_dir / "image" / "0000526-000071687000.jpg",
        gt_extract_dir / "image" / "0000549-000074827573.jpg",
        gt_extract_dir / "image" / "0000582-000079333613.jpg",
        gt_extract_dir / "image" / "0000630-000085887853.jpg",
        gt_extract_dir / "image" / "0000655-000089301520.jpg",
        gt_extract_dir / "image" / "0000703-000095855760.jpg",
        gt_extract_dir / "image" / "0000722-000098450147.jpg",
        gt_extract_dir / "image" / "0000771-000105140933.jpg",
        gt_extract_dir / "image" / "0000792-000108008413.jpg",
        gt_extract_dir / "image" / "0000818-000111558627.jpg",
        gt_extract_dir / "image" / "0000849-000115791573.jpg",
        gt_extract_dir / "image" / "0000883-000120434160.jpg",
        gt_extract_dir / "image" / "0000896-000122209267.jpg",
        gt_extract_dir / "image" / "0000935-000127534587.jpg",
        gt_extract_dir / "image" / "0000985-000134361920.jpg",
        gt_extract_dir / "image" / "0001028-000140233427.jpg",
        gt_extract_dir / "image" / "0001061-000144739467.jpg"
    ]
    assert len(fountain_dataset.color_paths) == len(gt_color_paths)
    for gt_path, fountain_dataset_color_path in zip(
            gt_color_paths, fountain_dataset.color_paths):
        assert Path(fountain_dataset_color_path) == gt_path
        assert Path(gt_path).is_file()

    gt_depth_paths = [
        gt_extract_dir / "depth" / "0000038-000001234662.png",
        gt_extract_dir / "depth" / "0000124-000004104418.png",
        gt_extract_dir / "depth" / "0000177-000005872988.png",
        gt_extract_dir / "depth" / "0000259-000008609267.png",
        gt_extract_dir / "depth" / "0000447-000014882686.png",
        gt_extract_dir / "depth" / "0000635-000021156105.png",
        gt_extract_dir / "depth" / "0000815-000027162570.png",
        gt_extract_dir / "depth" / "0000877-000029231463.png",
        gt_extract_dir / "depth" / "0001040-000034670651.png",
        gt_extract_dir / "depth" / "0001220-000040677116.png",
        gt_extract_dir / "depth" / "0001351-000045048488.png",
        gt_extract_dir / "depth" / "0001503-000050120614.png",
        gt_extract_dir / "depth" / "0001683-000056127079.png",
        gt_extract_dir / "depth" / "0001752-000058429557.png",
        gt_extract_dir / "depth" / "0001937-000064602868.png",
        gt_extract_dir / "depth" / "0001990-000066371438.png",
        gt_extract_dir / "depth" / "0002149-000071677149.png",
        gt_extract_dir / "depth" / "0002243-000074813859.png",
        gt_extract_dir / "depth" / "0002378-000079318707.png",
        gt_extract_dir / "depth" / "0002575-000085892450.png",
        gt_extract_dir / "depth" / "0002677-000089296113.png",
        gt_extract_dir / "depth" / "0002874-000095869855.png",
        gt_extract_dir / "depth" / "0002951-000098439288.png",
        gt_extract_dir / "depth" / "0003152-000105146507.png",
        gt_extract_dir / "depth" / "0003238-000108016262.png",
        gt_extract_dir / "depth" / "0003344-000111553403.png",
        gt_extract_dir / "depth" / "0003471-000115791298.png",
        gt_extract_dir / "depth" / "0003610-000120429623.png",
        gt_extract_dir / "depth" / "0003663-000122198194.png",
        gt_extract_dir / "depth" / "0003823-000127537274.png",
        gt_extract_dir / "depth" / "0004028-000134377970.png",
        gt_extract_dir / "depth" / "0004203-000140217589.png",
        gt_extract_dir / "depth" / "0004339-000144755807.png"
    ]
    assert len(fountain_dataset.depth_paths) == len(gt_depth_paths)
    for gt_path, fountain_dataset_depth_path in zip(
            gt_depth_paths, fountain_dataset.depth_paths):
        assert Path(fountain_dataset_depth_path) == gt_path
        assert Path(gt_path).is_file()

    assert Path(fountain_dataset.keyframe_poses_log_path
               ) == gt_extract_dir / "scene" / "key.log"
    assert Path(fountain_dataset.reconstruction_path
               ) == gt_extract_dir / "scene" / "integrated.ply"

    assert fountain_dataset.prefix == gt_prefix
    assert Path(fountain_dataset.data_root) == gt_data_root
    assert Path(fountain_dataset.download_dir) == gt_download_dir
    assert Path(fountain_dataset.extract_dir) == gt_extract_dir


def test_eagle():
    gt_prefix = "EaglePointCloud"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    eagle = o3d.data.EaglePointCloud()
    assert Path(gt_download_dir).is_dir()

    assert Path(eagle.path) == gt_extract_dir / "EaglePointCloud.ply"
    assert Path(eagle.path).is_file()

    assert eagle.prefix == gt_prefix
    assert Path(eagle.data_root) == gt_data_root
    assert Path(eagle.download_dir) == gt_download_dir
    assert Path(eagle.extract_dir) == gt_extract_dir


def test_armadillo():
    gt_prefix = "ArmadilloMesh"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    armadillo = o3d.data.ArmadilloMesh()
    assert Path(gt_download_dir).is_dir()

    assert Path(armadillo.path) == gt_extract_dir / "ArmadilloMesh.ply"
    assert Path(armadillo.path).is_file()

    assert armadillo.prefix == gt_prefix
    assert Path(armadillo.data_root) == gt_data_root
    assert Path(armadillo.download_dir) == gt_download_dir
    assert Path(armadillo.extract_dir) == gt_extract_dir


def test_bunny():
    gt_prefix = "BunnyMesh"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    bunny = o3d.data.BunnyMesh()
    assert Path(gt_download_dir).is_dir()

    assert Path(bunny.path) == gt_extract_dir / "BunnyMesh.ply"
    assert Path(bunny.path).is_file()

    assert bunny.prefix == gt_prefix
    assert Path(bunny.data_root) == gt_data_root
    assert Path(bunny.download_dir) == gt_download_dir
    assert Path(bunny.extract_dir) == gt_extract_dir


def test_knot():
    gt_prefix = "KnotMesh"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    knot = o3d.data.KnotMesh()
    assert Path(gt_download_dir).is_dir()

    assert Path(knot.path) == gt_extract_dir / "KnotMesh.ply"
    assert Path(knot.path).is_file()

    assert knot.prefix == gt_prefix
    assert Path(knot.data_root) == gt_data_root
    assert Path(knot.download_dir) == gt_download_dir
    assert Path(knot.extract_dir) == gt_extract_dir


def test_monkey():
    gt_prefix = "MonkeyModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    monkey = o3d.data.MonkeyModel()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "albedo.png",
        "ao": Path(gt_extract_dir) / "ao.png",
        "metallic": Path(gt_extract_dir) / "metallic.png",
        "monkey_material": Path(gt_extract_dir) / "monkey.mtl",
        "monkey_model": Path(gt_extract_dir) / "monkey.obj",
        "monkey_solid_material": Path(gt_extract_dir) / "monkey_solid.mtl",
        "monkey_solid_model": Path(gt_extract_dir) / "monkey_solid.obj",
        "normal": Path(gt_extract_dir) / "normal.png",
        "roughness": Path(gt_extract_dir) / "roughness.png"
    }

    for file_name in monkey.path_map:
        assert Path(monkey.path_map[file_name]) == gt_path_map[file_name]
        assert Path(monkey.path_map[file_name]).is_file()

    assert Path(monkey.path) == gt_extract_dir / "monkey.obj"
    assert Path(monkey.path).is_file()

    assert monkey.prefix == gt_prefix
    assert Path(monkey.data_root) == gt_data_root
    assert Path(monkey.download_dir) == gt_download_dir
    assert Path(monkey.extract_dir) == gt_extract_dir


def test_sword():
    gt_prefix = "SwordModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    sword = o3d.data.SwordModel()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "sword_material": Path(gt_extract_dir) / "UV.mtl",
        "sword_model": Path(gt_extract_dir) / "UV.obj",
        "base_color": Path(gt_extract_dir) / "UV_blinn1SG_BaseColor.png",
        "metallic": Path(gt_extract_dir) / "UV_blinn1SG_Metallic.png",
        "normal": Path(gt_extract_dir) / "UV_blinn1SG_Normal.png",
        "roughness": Path(gt_extract_dir) / "UV_blinn1SG_Roughness.png"
    }

    for file_name in sword.path_map:
        assert Path(sword.path_map[file_name]) == gt_path_map[file_name]
        assert Path(sword.path_map[file_name]).is_file()

    assert Path(sword.path) == gt_extract_dir / "UV.obj"
    assert Path(sword.path).is_file()

    assert sword.prefix == gt_prefix
    assert Path(sword.data_root) == gt_data_root
    assert Path(sword.download_dir) == gt_download_dir
    assert Path(sword.extract_dir) == gt_extract_dir


def test_crate():
    gt_prefix = "CrateModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    crate = o3d.data.CrateModel()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "crate_material": Path(gt_extract_dir) / "crate.mtl",
        "crate_model": Path(gt_extract_dir) / "crate.obj",
        "texture_image": Path(gt_extract_dir) / "crate.jpg"
    }

    for file_name in crate.path_map:
        assert Path(crate.path_map[file_name]) == gt_path_map[file_name]
        assert Path(crate.path_map[file_name]).is_file()

    assert Path(crate.path) == gt_extract_dir / "crate.obj"
    assert Path(crate.path).is_file()

    assert crate.prefix == gt_prefix
    assert Path(crate.data_root) == gt_data_root
    assert Path(crate.download_dir) == gt_download_dir
    assert Path(crate.extract_dir) == gt_extract_dir


def test_flight_helmet():
    gt_prefix = "FlightHelmetModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    helmet = o3d.data.FlightHelmetModel()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "flight_helmet":
            Path(gt_extract_dir) / "FlightHelmet.gltf",
        "flight_helmet_bin":
            Path(gt_extract_dir) / "FlightHelmet.bin",
        "mat_glass_plastic_base":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_GlassPlasticMat_BaseColor.png",
        "mat_glass_plastic_normal":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_GlassPlasticMat_Normal.png",
        "mat_glass_plastic_occlusion_rough_metal":
            Path(gt_extract_dir) / "FlightHelmet_Materials_GlassPlasticMat_"
            "OcclusionRoughMetal.png",
        "mat_leather_parts_base":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_LeatherPartsMat_BaseColor.png",
        "mat_leather_parts_normal":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_LeatherPartsMat_Normal.png",
        "mat_leather_parts_occlusion_rough_metal":
            Path(gt_extract_dir) / "FlightHelmet_Materials_LeatherPartsMat_"
            "OcclusionRoughMetal.png",
        "mat_lenses_base":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_LensesMat_BaseColor.png",
        "mat_lenses_normal":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_LensesMat_Normal.png",
        "mat_lenses_occlusion_rough_metal":
            Path(gt_extract_dir) / "FlightHelmet_Materials_LensesMat_"
            "OcclusionRoughMetal.png",
        "mat_metal_parts_base":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_MetalPartsMat_BaseColor.png",
        "mat_metal_parts_normal":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_MetalPartsMat_Normal.png",
        "mat_metal_parts_occlusion_rough_metal":
            Path(gt_extract_dir) / "FlightHelmet_Materials_MetalPartsMat_"
            "OcclusionRoughMetal.png",
        "mat_rubber_wood_base":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_RubberWoodMat_BaseColor.png",
        "mat_rubber_wood_normal":
            Path(gt_extract_dir) /
            "FlightHelmet_Materials_RubberWoodMat_Normal.png",
        "mat_rubber_wood_occlusion_rough_metal":
            Path(gt_extract_dir) / "FlightHelmet_Materials_RubberWoodMat_"
            "OcclusionRoughMetal.png"
    }

    for file_name in helmet.path_map:
        assert Path(helmet.path_map[file_name]) == gt_path_map[file_name]
        assert Path(helmet.path_map[file_name]).is_file()

    assert Path(helmet.path) == gt_extract_dir / "FlightHelmet.gltf"
    assert Path(helmet.path).is_file()

    assert helmet.prefix == gt_prefix
    assert Path(helmet.data_root) == gt_data_root
    assert Path(helmet.download_dir) == gt_download_dir
    assert Path(helmet.extract_dir) == gt_extract_dir


def test_avocado():
    gt_prefix = "AvocadoModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    avocado = o3d.data.AvocadoModel()
    assert Path(gt_download_dir).is_dir()

    assert Path(avocado.path) == gt_extract_dir / "AvocadoModel.glb"
    assert Path(avocado.path).is_file()

    assert avocado.prefix == gt_prefix
    assert Path(avocado.data_root) == gt_data_root
    assert Path(avocado.download_dir) == gt_download_dir
    assert Path(avocado.extract_dir) == gt_extract_dir


def test_damaged_helmet():
    gt_prefix = "DamagedHelmetModel"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    damaged_helmet = o3d.data.DamagedHelmetModel()
    assert Path(gt_download_dir).is_dir()

    assert Path(
        damaged_helmet.path) == gt_extract_dir / "DamagedHelmetModel.glb"
    assert Path(damaged_helmet.path).is_file()

    assert damaged_helmet.prefix == gt_prefix
    assert Path(damaged_helmet.data_root) == gt_data_root
    assert Path(damaged_helmet.download_dir) == gt_download_dir
    assert Path(damaged_helmet.extract_dir) == gt_extract_dir


def test_metal_texture():
    gt_prefix = "MetalTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.MetalTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "Metal008_Color.jpg",
        "normal": Path(gt_extract_dir) / "Metal008_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "Metal008_Roughness.jpg",
        "metallic": Path(gt_extract_dir) / "Metal008_Metalness.jpg"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]
    assert Path(data.metallic_texture_path) == gt_path_map["metallic"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_painted_plaster_texture():
    gt_prefix = "PaintedPlasterTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.PaintedPlasterTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "PaintedPlaster017_Color.jpg",
        "normal": Path(gt_extract_dir) / "PaintedPlaster017_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "noiseTexture.png"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_tiles_texture():
    gt_prefix = "TilesTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.TilesTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "Tiles074_Color.jpg",
        "normal": Path(gt_extract_dir) / "Tiles074_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "Tiles074_Roughness.jpg"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_terrazzo_texture():
    gt_prefix = "TerrazzoTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.TerrazzoTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "Terrazzo018_Color.jpg",
        "normal": Path(gt_extract_dir) / "Terrazzo018_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "Terrazzo018_Roughness.jpg"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_wood_texture():
    gt_prefix = "WoodTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.WoodTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "Wood049_Color.jpg",
        "normal": Path(gt_extract_dir) / "Wood049_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "Wood049_Roughness.jpg"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_wood_floor_texture():
    gt_prefix = "WoodFloorTexture"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    data = o3d.data.WoodFloorTexture()
    assert Path(gt_download_dir).is_dir()

    gt_path_map = {
        "albedo": Path(gt_extract_dir) / "WoodFloor050_Color.jpg",
        "normal": Path(gt_extract_dir) / "WoodFloor050_NormalDX.jpg",
        "roughness": Path(gt_extract_dir) / "WoodFloor050_Roughness.jpg"
    }

    for file_name in data.path_map:
        assert Path(data.path_map[file_name]) == gt_path_map[file_name]
        assert Path(data.path_map[file_name]).is_file()

    assert Path(data.albedo_texture_path) == gt_path_map["albedo"]
    assert Path(data.normal_texture_path) == gt_path_map["normal"]
    assert Path(data.roughness_texture_path) == gt_path_map["roughness"]

    assert data.prefix == gt_prefix
    assert Path(data.data_root) == gt_data_root
    assert Path(data.download_dir) == gt_download_dir
    assert Path(data.extract_dir) == gt_extract_dir


def test_juneau():
    gt_prefix = "JuneauImage"
    gt_data_root, gt_download_dir, gt_extract_dir = get_test_data_dirs(
        gt_prefix)

    juneau = o3d.data.JuneauImage()
    assert Path(gt_download_dir).is_dir()

    assert Path(juneau.path) == gt_extract_dir / "JuneauImage.jpg"
    assert Path(juneau.path).is_file()

    assert juneau.prefix == gt_prefix
    assert Path(juneau.data_root) == gt_data_root
    assert Path(juneau.download_dir) == gt_download_dir
    assert Path(juneau.extract_dir) == gt_extract_dir


@pytest.mark.skip()
def test_redwood_indoor_living_room1():
    gt_prefix = "RedwoodIndoorLivingRoom1"
    _, gt_download_dir, gt_extract_dir = get_test_data_dirs(gt_prefix)
    dataset = o3d.data.RedwoodIndoorLivingRoom1()

    assert Path(gt_download_dir).is_dir()
    assert Path(gt_extract_dir).is_dir()

    pcd = o3d.io.read_point_cloud(dataset.point_cloud_path)

    im_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths, dataset.depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_rgbds.append(im_rgbd)
    assert len(im_rgbds) == 2870

    im_noisy_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths,
                                      dataset.noisy_depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_noisy_rgbds.append(im_rgbd)
    assert len(im_noisy_rgbds) == 2870


@pytest.mark.skip()
def test_redwood_indoor_living_room2():
    gt_prefix = "RedwoodIndoorLivingRoom2"
    _, gt_download_dir, gt_extract_dir = get_test_data_dirs(gt_prefix)
    dataset = o3d.data.RedwoodIndoorLivingRoom2()

    assert Path(gt_download_dir).is_dir()
    assert Path(gt_extract_dir).is_dir()

    pcd = o3d.io.read_point_cloud(dataset.point_cloud_path)

    im_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths, dataset.depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_rgbds.append(im_rgbd)
    assert len(im_rgbds) == 2350

    im_noisy_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths,
                                      dataset.noisy_depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_noisy_rgbds.append(im_rgbd)
    assert len(im_noisy_rgbds) == 2350


@pytest.mark.skip()
def test_redwood_indoor_office1():
    gt_prefix = "RedwoodIndoorOffice1"
    _, gt_download_dir, gt_extract_dir = get_test_data_dirs(gt_prefix)
    dataset = o3d.data.RedwoodIndoorOffice1()

    assert Path(gt_download_dir).is_dir()
    assert Path(gt_extract_dir).is_dir()

    pcd = o3d.io.read_point_cloud(dataset.point_cloud_path)

    im_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths, dataset.depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_rgbds.append(im_rgbd)
    assert len(im_rgbds) == 2690

    im_noisy_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths,
                                      dataset.noisy_depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_noisy_rgbds.append(im_rgbd)
    assert len(im_noisy_rgbds) == 2690


@pytest.mark.skip()
def test_redwood_indoor_office2():
    gt_prefix = "RedwoodIndoorOffice2"
    _, gt_download_dir, gt_extract_dir = get_test_data_dirs(gt_prefix)
    dataset = o3d.data.RedwoodIndoorOffice2()

    assert Path(gt_download_dir).is_dir()
    assert Path(gt_extract_dir).is_dir()

    pcd = o3d.io.read_point_cloud(dataset.point_cloud_path)

    im_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths, dataset.depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_rgbds.append(im_rgbd)
    assert len(im_rgbds) == 2538

    im_noisy_rgbds = []
    for color_path, depth_path in zip(dataset.color_paths,
                                      dataset.noisy_depth_paths):
        im_color = o3d.io.read_image(color_path)
        im_depth = o3d.io.read_image(depth_path)
        im_rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            im_color, im_depth)
        im_noisy_rgbds.append(im_rgbd)
    assert len(im_noisy_rgbds) == 2538



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import os
import platform
import subprocess

import numpy as np
import tabulate
import nvidia_smi


def get_processor_name():
    if platform.system() == "Windows":
        return platform.processor()
    elif platform.system() == "Darwin":
        os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
        command = "sysctl -n machdep.cpu.brand_string"
        return subprocess.check_output(command).strip()
    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo | grep 'model name' -m 1"
        name = subprocess.check_output(command, shell=True).strip()
        return str(name, 'utf-8')


def run_command(command):
    result = subprocess.run(command,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True,
                            shell=True)
    return result.stdout.strip()


def print_system_info():
    # print results
    nvcc_version = run_command("nvcc --version")
    os_version = run_command("cat /etc/os-release")
    cpu_info = get_processor_name()
    gpu_info = run_command("nvidia-smi")
    print("======System Info======")
    print("[CUDA]")
    print(nvcc_version)
    print("")

    print("[OS]")
    print(os_version)
    print("")

    print("[CPU]")
    print(cpu_info)
    print("")

    print("[GPU]")
    print(gpu_info)
    print("======System Info [End]=")


def measure_time(fn, min_samples=10, max_samples=100, max_time_in_sec=10.0):
    """Measure time to run fn. Returns the elapsed time each run."""
    from time import perf_counter_ns
    t = []
    for i in range(max_samples):
        if sum(t) / 1e9 >= max_time_in_sec and i >= min_samples:
            break
        t.append(-perf_counter_ns())
        try:
            ans = fn()
        except Exception as e:
            print(e)
            return np.array([np.nan])
        t[-1] += perf_counter_ns()
        del ans
    print('.', end='')
    return np.array(t) / 1e9


def measure_memory(fn, handle):
    """Measure memory to run fn. Returns the maximum allocated memory each run."""
    try:
        _ = fn()
    except Exception as e:
        print(e)
        return np.nan
    info = nvidia_smi.nvmlDeviceGetMemoryInfo(handle)
    memory = info.used / 1000. / 1000. / 1000.
    print('.', end='')
    return memory


def print_table(methods, results):
    headers = [''] + [f'{n}_setup' for n in methods
                     ] + [f'{n}_search' for n in methods]
    rows = []

    for x in results[0]:
        r = [x] + list(
            map(np.median, [r[x]['setup'] for r in results] +
                [r[x]['search'] for r in results]))
        rows.append(r)

    print(tabulate.tabulate(rows, headers=headers))


def print_table_simple(methods, results):
    headers = [''] + [f'{n}_search' for n in methods]
    rows = []

    for x in results[0]:
        r = [x] + list(map(np.median, [r[x]['search'] for r in results]))
        rows.append(r)

    print(tabulate.tabulate(rows, headers=headers))


def print_table_memory(methods, results):
    headers = [''] + [f'{n}' for n in methods]
    rows = []

    for x in results[0]:
        r = [x] + list(map(np.median, [r[x]['memory'] for r in results]))
        rows.append(r)

    print(tabulate.tabulate(rows, headers=headers))


def sample_points(points, num_sample):
    if points.shape[0] < num_sample:
        num_sample = points.shape[0]

    idx = np.round(np.linspace(0, len(points) - 1, num_sample)).astype(int)
    return points[idx]



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

# examples/python/reconstruction_system/debug/pairwise_pc_alignment.py

import argparse
import json
import sys

pyexample_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(pyexample_path)
from open3d_example import *

sys.path.append(".")
from initialize_config import *
from register_fragments import *

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="matching two point clouds")
    parser.add_argument("config", help="reading json file for initial pose")
    parser.add_argument("source_id", type=int, help="ID of source point cloud")
    parser.add_argument("target_id", type=int, help="ID of target point cloud")
    args = parser.parse_args()

    with open(args.config) as json_file:
        config = json.load(json_file)
        initialize_config(config)
        config['debug_mode'] = True
        ply_file_names = get_file_list(
            join(config["path_dataset"], config["folder_fragment"]), ".ply")
        register_point_cloud_pair(ply_file_names, args.source_id,
                                  args.target_id, config)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import numpy as np
import open3d as o3d
import threading
import time

CLOUD_NAME = "points"


def main():
    MultiWinApp().run()


class MultiWinApp:

    def __init__(self):
        self.is_done = False
        self.n_snapshots = 0
        self.cloud = None
        self.main_vis = None
        self.snapshot_pos = None

    def run(self):
        app = o3d.visualization.gui.Application.instance
        app.initialize()

        self.main_vis = o3d.visualization.O3DVisualizer(
            "Open3D - Multi-Window Demo")
        self.main_vis.add_action("Take snapshot in new window",
                                 self.on_snapshot)
        self.main_vis.set_on_close(self.on_main_window_closing)

        app.add_window(self.main_vis)
        self.snapshot_pos = (self.main_vis.os_frame.x, self.main_vis.os_frame.y)

        threading.Thread(target=self.update_thread).start()

        app.run()

    def on_snapshot(self, vis):
        self.n_snapshots += 1
        self.snapshot_pos = (self.snapshot_pos[0] + 50,
                             self.snapshot_pos[1] + 50)
        title = "Open3D - Multi-Window Demo (Snapshot #" + str(
            self.n_snapshots) + ")"
        new_vis = o3d.visualization.O3DVisualizer(title)
        mat = o3d.visualization.rendering.MaterialRecord()
        mat.shader = "defaultUnlit"
        new_vis.add_geometry(CLOUD_NAME + " #" + str(self.n_snapshots),
                             self.cloud, mat)
        new_vis.reset_camera_to_default()
        bounds = self.cloud.get_axis_aligned_bounding_box()
        extent = bounds.get_extent()
        new_vis.setup_camera(60, bounds.get_center(),
                             bounds.get_center() + [0, 0, -3], [0, -1, 0])
        o3d.visualization.gui.Application.instance.add_window(new_vis)
        new_vis.os_frame = o3d.visualization.gui.Rect(self.snapshot_pos[0],
                                                      self.snapshot_pos[1],
                                                      new_vis.os_frame.width,
                                                      new_vis.os_frame.height)

    def on_main_window_closing(self):
        self.is_done = True
        return True  # False would cancel the close

    def update_thread(self):
        # This is NOT the UI thread, need to call post_to_main_thread() to update
        # the scene or any part of the UI.
        pcd_data = o3d.data.DemoICPPointClouds()
        self.cloud = o3d.io.read_point_cloud(pcd_data.paths[0])
        bounds = self.cloud.get_axis_aligned_bounding_box()
        extent = bounds.get_extent()

        def add_first_cloud():
            mat = o3d.visualization.rendering.MaterialRecord()
            mat.shader = "defaultUnlit"
            self.main_vis.add_geometry(CLOUD_NAME, self.cloud, mat)
            self.main_vis.reset_camera_to_default()
            self.main_vis.setup_camera(60, bounds.get_center(),
                                       bounds.get_center() + [0, 0, -3],
                                       [0, -1, 0])

        o3d.visualization.gui.Application.instance.post_to_main_thread(
            self.main_vis, add_first_cloud)

        while not self.is_done:
            time.sleep(0.1)

            # Perturb the cloud with a random walk to simulate an actual read
            pts = np.asarray(self.cloud.points)
            magnitude = 0.005 * extent
            displacement = magnitude * (np.random.random_sample(pts.shape) -
                                        0.5)
            new_pts = pts + displacement
            self.cloud.points = o3d.utility.Vector3dVector(new_pts)

            def update_cloud():
                # Note: if the number of points is less than or equal to the
                #       number of points in the original object that was added,
                #       using self.scene.update_geometry() will be faster.
                #       Requires that the point cloud be a t.PointCloud.
                self.main_vis.remove_geometry(CLOUD_NAME)
                mat = o3d.visualization.rendering.MaterialRecord()
                mat.shader = "defaultUnlit"
                self.main_vis.add_geometry(CLOUD_NAME, self.cloud, mat)

            if self.is_done:  # might have changed while sleeping
                break
            o3d.visualization.gui.Application.instance.post_to_main_thread(
                self.main_vis, update_cloud)


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d

if __name__ == "__main__":
    sample_ply_data = o3d.data.PLYPointCloud()
    pcd = o3d.io.read_point_cloud(sample_ply_data.path)
    # Flip it, otherwise the pointcloud will be upside down.
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    print(pcd)
    axis_aligned_bounding_box = pcd.get_axis_aligned_bounding_box()
    axis_aligned_bounding_box.color = (1, 0, 0)
    oriented_bounding_box = pcd.get_oriented_bounding_box()
    oriented_bounding_box.color = (0, 1, 0)
    print(
        "Displaying axis_aligned_bounding_box in red and oriented bounding box in green ..."
    )
    o3d.visualization.draw(
        [pcd, axis_aligned_bounding_box, oriented_bounding_box])



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import numpy as np
import pytest
from open3d.ml.contrib import subsample, subsample_batch


def compare_results_with_sorting(actual_points,
                                 expect_points,
                                 actual_features=None,
                                 expect_features=None,
                                 actual_labels=None,
                                 expect_labels=None):
    actual_argsort = actual_points.argsort(axis=0)[:, 0]
    expect_argsort = expect_points.argsort(axis=0)[:, 0]
    np.testing.assert_allclose(actual_points[actual_argsort],
                               expect_points[expect_argsort])
    if actual_features is not None and expect_features is not None:
        np.testing.assert_allclose(actual_features[actual_argsort],
                                   expect_features[expect_argsort])
    if actual_labels is not None and expect_labels is not None:
        np.testing.assert_equal(actual_labels[actual_argsort],
                                expect_labels[expect_argsort])


def compare_batched_results_with_sorting(actual_points,
                                         expect_points,
                                         actual_batches,
                                         expect_batches,
                                         actual_features=None,
                                         expect_features=None,
                                         actual_labels=None,
                                         expect_labels=None):
    np.testing.assert_equal(actual_batches, expect_batches)
    assert actual_batches.ndim == 1
    end_indices = np.cumsum(actual_batches)
    start_indices = [0] + list(end_indices[:-1])
    for s, e in zip(start_indices, end_indices):
        compare_results_with_sorting(actual_points=actual_points[s:e],
                                     expect_points=expect_points[s:e],
                                     actual_features=actual_features[s:e]
                                     if actual_features is not None else None,
                                     expect_features=expect_features[s:e]
                                     if expect_features is not None else None,
                                     actual_labels=actual_labels[s:e]
                                     if actual_labels is not None else None,
                                     expect_labels=expect_labels[s:e]
                                     if expect_labels is not None else None)


def test_subsample():
    points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1],
                       [5, 0, 0], [5, 1, 0]],
                      dtype=np.float32)
    features = np.array(range(21), dtype=np.float32).reshape(-1, 3)
    labels = np.array([0, 0, 0, 0, 0, 1, 1], dtype=np.int32)

    # Reference results.
    sub_points_ref = np.array([[5, 0.5, 0], [0.4, 0.4, 0.4]], dtype=np.float32)
    sub_features_ref = np.array([[16.5, 17.5, 18.5], [6, 7, 8]],
                                dtype=np.float32)
    sub_labels_ref = np.array([1, 0], dtype=np.int32)

    # Passing only points.
    sub_points = subsample(points, sampleDl=1.1)
    compare_results_with_sorting(actual_points=sub_points,
                                 expect_points=sub_points_ref)

    # Passing points and features.
    sub_points, sub_features = subsample(points,
                                         features=features,
                                         sampleDl=1.1)
    compare_results_with_sorting(actual_points=sub_points,
                                 expect_points=sub_points_ref,
                                 actual_features=sub_features,
                                 expect_features=sub_features_ref)

    # Passing points, features and labels.
    sub_points, sub_features, sub_labels = subsample(points,
                                                     features=features,
                                                     classes=labels,
                                                     sampleDl=1.1)
    compare_results_with_sorting(actual_points=sub_points,
                                 expect_points=sub_points_ref,
                                 actual_features=sub_features,
                                 expect_features=sub_features_ref,
                                 actual_labels=sub_labels,
                                 expect_labels=sub_labels_ref)

    # Test wrong dtype.
    with pytest.raises(RuntimeError):
        sub_points = subsample(np.array(points, dtype=np.int32), sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(np.array(points, dtype=np.float64), sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(points,
                               features=np.array(features, dtype=np.int32),
                               sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(points,
                               features=features,
                               classes=np.array(labels, dtype=np.float32),
                               sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(points,
                               features=np.array(features, np.int32),
                               classes=np.array(labels, dtype=np.float32),
                               sampleDl=1.1)

    # Test shape mismatch
    with pytest.raises(RuntimeError):
        sub_points = subsample(points[0], sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(np.ones((10, 4), dtype=np.float32), sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(points, features=features[0], sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample(points[0], sampleDl=1.1)
    with pytest.raises(TypeError):
        sub_points = subsample(None, sampleDl=1.1)


def test_subsample_batch():
    points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1],
                       [5, 0, 0], [5, 1, 0]],
                      dtype=np.float32)
    features = np.array(range(28), dtype=np.float32).reshape(-1, 4)
    labels = np.array([0, 0, 3, 1, 1, 2, 2], dtype=np.int32)
    batches = np.array([3, 2, 2], dtype=np.int32)

    # Reference results.
    sub_points_ref = np.array(
        [[0.3333333, 0.3333333, 0], [0.5, 0.5, 1], [5, 0.5, 0]],
        dtype=np.float32)
    sub_batch_ref = np.array([1, 1, 1], dtype=np.int32)
    sub_labels_ref = np.array([0, 1, 2], dtype=np.int32)

    # Passing only points.
    sub_points, sub_batch = subsample_batch(points, batches, sampleDl=1.1)
    compare_batched_results_with_sorting(actual_points=sub_points,
                                         expect_points=sub_points_ref,
                                         actual_batches=sub_batch,
                                         expect_batches=sub_batch_ref)

    # Passing points and features.
    sub_features_ref = np.array(
        [[4, 5, 6, 7], [14, 15, 16, 17], [22, 23, 24, 25]], dtype=np.float32)
    sub_points, sub_batch, sub_features = subsample_batch(points,
                                                          batches,
                                                          features=features,
                                                          sampleDl=1.1)
    compare_batched_results_with_sorting(actual_points=sub_points,
                                         expect_points=sub_points_ref,
                                         actual_batches=sub_batch,
                                         expect_batches=sub_batch_ref,
                                         actual_features=sub_features,
                                         expect_features=sub_features_ref)

    # Passing points, features and labels.
    sub_points, sub_batch, sub_features, sub_labels = subsample_batch(
        points, batches, features=features, classes=labels, sampleDl=1.1)
    compare_batched_results_with_sorting(actual_points=sub_points,
                                         expect_points=sub_points_ref,
                                         actual_batches=sub_batch,
                                         expect_batches=sub_batch_ref,
                                         actual_features=sub_features,
                                         expect_features=sub_features_ref,
                                         actual_labels=sub_labels,
                                         expect_labels=sub_labels_ref)

    # Test wrong dtype.
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(np.array(points, dtype=np.int32),
                                     batches,
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     np.array(batches, dtype=np.float32),
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(np.array(points, dtype=np.float64),
                                     batches,
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     batches,
                                     features=np.array(features,
                                                       dtype=np.int32),
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     batches,
                                     features=features,
                                     classes=np.array(labels, dtype=np.float32),
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     batches,
                                     features=np.array(features, np.int32),
                                     classes=np.array(labels, dtype=np.float32),
                                     sampleDl=1.1)

    # Test shape mismatch
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points[0], batches, sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(np.ones((10, 4), dtype=np.float32),
                                     batches,
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     batches,
                                     features=features[0],
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points[0], batches, sampleDl=1.1)
    with pytest.raises(TypeError):
        sub_points = subsample_batch(None, None, sampleDl=1.1)

    # Test sum(batch) != num_points
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     np.array([3, 3, 2], dtype=np.int32),
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     np.array([3, 3, 2], dtype=np.int32),
                                     features=features,
                                     classes=labels,
                                     sampleDl=1.1)
    with pytest.raises(RuntimeError):
        sub_points = subsample_batch(points,
                                     np.array([1], dtype=np.int32),
                                     sampleDl=1.1)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------
"""This example shows Open3D's remote visualization feature using RPC
communication. To run this example, start the client first by running

python remote_visualizer.py client

and then run the server by running

python remote_visualizer.py server

Port 51454 is used by default for communication. For remote visualization (client
and server running on different machines), use ssh to forward the remote server
port to your local computer:

    ssh -N -R 51454:localhost:51454 user@remote_host

See documentation for more details (e.g. to use a different port).
"""
import sys
import numpy as np
import open3d as o3d
import open3d.visualization as vis


def make_point_cloud(npts, center, radius, colorize):
    pts = np.random.uniform(-radius, radius, size=[npts, 3]) + center
    cloud = o3d.geometry.PointCloud()
    cloud.points = o3d.utility.Vector3dVector(pts)
    if colorize:
        colors = np.random.uniform(0.0, 1.0, size=[npts, 3])
        cloud.colors = o3d.utility.Vector3dVector(colors)
    return cloud


def server_time_animation():
    orig = make_point_cloud(200, (0, 0, 0), 1.0, True)
    clouds = [{"name": "t=0", "geometry": orig, "time": 0}]
    drift_dir = (1.0, 0.0, 0.0)
    expand = 1.0
    n = 20
    ev = o3d.visualization.ExternalVisualizer()
    for i in range(1, n):
        amount = float(i) / float(n - 1)
        cloud = o3d.geometry.PointCloud()
        pts = np.asarray(orig.points)
        pts = pts * (1.0 + amount * expand) + [amount * v for v in drift_dir]
        cloud.points = o3d.utility.Vector3dVector(pts)
        cloud.colors = orig.colors
        ev.set(obj=cloud, time=i, path=f"points at t={i}")
        print('.', end='', flush=True)
    print()


def client_time_animation():
    o3d.visualization.draw(title="Open3D - Remote Visualizer Client",
                           show_ui=True,
                           rpc_interface=True)


if __name__ == "__main__":
    assert len(sys.argv) == 2 and sys.argv[1] in ('client', 'server'), (
        "Usage: python remote_visualizer.py [client|server]")
    if sys.argv[1] == "client":
        client_time_animation()
    elif sys.argv[1] == "server":
        server_time_animation()



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------
"""ICP variant that uses both geometry and color for registration"""

import open3d as o3d
import numpy as np
import copy


def draw_registration_result(source, target, transformation):
    source_temp = copy.deepcopy(source)
    source_temp.transform(transformation)
    o3d.visualization.draw([source_temp, target])


print("Load two point clouds and show initial pose ...")
ply_data = o3d.data.DemoColoredICPPointClouds()
source = o3d.io.read_point_cloud(ply_data.paths[0])
target = o3d.io.read_point_cloud(ply_data.paths[1])

if __name__ == "__main__":
    # Draw initial alignment.
    current_transformation = np.identity(4)
    # draw_registration_result(source, target, current_transformation)
    print(current_transformation)

    # Colored pointcloud registration.
    # This is implementation of following paper:
    # J. Park, Q.-Y. Zhou, V. Koltun,
    # Colored Point Cloud Registration Revisited, ICCV 2017.
    voxel_radius = [0.04, 0.02, 0.01]
    max_iter = [50, 30, 14]
    current_transformation = np.identity(4)
    print("Colored point cloud registration ...\n")
    for scale in range(3):
        iter = max_iter[scale]
        radius = voxel_radius[scale]
        print([iter, radius, scale])

        print("1. Downsample with a voxel size %.2f" % radius)
        source_down = source.voxel_down_sample(radius)
        target_down = target.voxel_down_sample(radius)

        print("2. Estimate normal")
        source_down.estimate_normals(
            o3d.geometry.KDTreeSearchParamHybrid(radius=radius * 2, max_nn=30))
        target_down.estimate_normals(
            o3d.geometry.KDTreeSearchParamHybrid(radius=radius * 2, max_nn=30))

        print("3. Applying colored point cloud registration")
        result_icp = o3d.pipelines.registration.registration_colored_icp(
            source_down, target_down, radius, current_transformation,
            o3d.pipelines.registration.TransformationEstimationForColoredICP(),
            o3d.pipelines.registration.ICPConvergenceCriteria(
                relative_fitness=1e-6, relative_rmse=1e-6, max_iteration=iter))
        current_transformation = result_icp.transformation
        print(result_icp, "\n")
    # draw_registration_result(source, target, result_icp.transformation)
    print(current_transformation)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------
"""This module loads the op library."""

import os as _os
import sys as _sys
import tensorflow as _tf
from open3d import _build_config

_lib_path = []
# allow overriding the path to the op library with an env var.
if 'OPEN3D_TF_OP_LIB' in _os.environ:
    _lib_path.append(_os.environ['OPEN3D_TF_OP_LIB'])

_this_dir = _os.path.dirname(__file__)
_package_root = _os.path.join(_this_dir, '..', '..', '..', '..')
_lib_ext = {'linux': '.so', 'darwin': '.dylib', 'win32': '.dll'}[_sys.platform]
_lib_suffix = '_debug' if _build_config['CMAKE_BUILD_TYPE'] == 'Debug' else ''
_lib_arch = ('cuda', 'cpu') if _build_config["BUILD_CUDA_MODULE"] else ('cpu',)
_lib_path.extend([
    _os.path.join(_package_root, la, 'open3d_tf_ops' + _lib_suffix + _lib_ext)
    for la in _lib_arch
])

_load_except = None
_loaded = False
for _lp in _lib_path:
    try:
        _lib = _tf.load_op_library(_lp)
        _loaded = True
        break
    except Exception as ex:
        _load_except = ex
        if not _os.path.isfile(_lp):
            print('The op library at "{}" was not found. Make sure that '
                  'BUILD_TENSORFLOW_OPS was enabled.'.format(
                      _os.path.realpath(_lp)))

if not _loaded:
    raise _load_except



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import open3d.core as o3c
import numpy as np
import pytest

from open3d_test import list_devices


def get_pcds(dtype, device):

    source_points = o3c.Tensor(
        [[1.0, 0.5, 2.0], [0.5, 0.5, 2.0], [0.5, 0.5, 2.5], [3.0, 1.0, 1.5],
         [3.5, 1.0, 1.0], [0.5, 1.0, 2.0], [1.5, 1.0, 1.5], [2.0, 1.0, 1.5],
         [1.0, 1.0, 2.0], [2.5, 1.0, 1.5], [3.0, 1.0, 1.0], [0.5, 1.0, 2.5],
         [1.0, 1.0, 1.0], [1.5, 1.0, 1.0], [1.0, 1.5, 1.0], [3.0, 1.5, 1.0],
         [3.5, 1.5, 1.0], [3.0, 1.5, 1.5], [0.5, 1.5, 1.5], [0.5, 1.5, 2.0],
         [1.0, 1.5, 2.0], [2.5, 1.5, 1.5], [1.5, 1.5, 1.0], [1.5, 1.5, 1.5],
         [2.0, 1.5, 1.5], [3.0, 1.5, 0.5], [2.5, 1.5, 1.0], [2.0, 1.5, 1.0],
         [3.0, 2.0, 0.5], [0.5, 2.0, 1.5], [3.0, 2.0, 1.0], [1.0, 2.0, 1.0],
         [2.0, 1.5, 0.5], [0.5, 2.0, 2.0], [2.5, 2.0, 1.0], [2.5, 2.0, 0.5],
         [2.0, 2.0, 0.5], [2.5, 1.5, 0.5], [3.0, 2.0, 1.5], [2.0, 2.0, 1.0],
         [1.0, 2.0, 2.0], [1.5, 2.0, 1.0], [1.5, 2.0, 1.5], [2.5, 2.0, 1.5],
         [2.0, 2.0, 1.5], [1.0, 2.0, 0.5], [0.5, 2.0, 1.0], [1.5, 2.0, 0.5],
         [1.0, 2.0, 1.5]], dtype, device)

    target_points = o3c.Tensor(
        [[1.5, 1.0, 1.5], [2.5, 1.0, 1.5], [1.5, 1.0, 1.0], [1.0, 1.0, 1.0],
         [2.0, 1.0, 1.5], [3.0, 1.0, 1.5], [1.0, 1.0, 0.5], [1.0, 1.5, 1.0],
         [1.0, 1.5, 0.5], [1.0, 1.0, 1.5], [3.0, 1.0, 2.0], [3.0, 1.5, 2.0],
         [3.0, 1.5, 1.5], [1.0, 1.5, 1.5], [1.5, 1.5, 1.5], [2.5, 1.5, 1.5],
         [2.0, 1.5, 1.5], [1.5, 1.5, 1.0], [2.5, 1.5, 2.0], [1.0, 2.0, 1.0],
         [1.0, 2.0, 0.5], [2.5, 1.5, 1.0], [3.0, 2.0, 1.5], [2.5, 2.0, 1.0],
         [2.5, 2.0, 1.5], [1.5, 2.0, 1.0], [2.0, 1.5, 1.0], [1.0, 2.0, 1.5],
         [2.0, 2.0, 1.0], [1.5, 2.0, 1.5], [1.5, 2.0, 0.5], [2.0, 2.0, 1.5],
         [2.0, 2.0, 0.5], [1.5, 2.5, 1.0], [1.0, 2.5, 1.0], [3.0, 2.0, 1.0],
         [2.0, 2.5, 1.0], [2.5, 2.5, 1.0]], dtype, device)

    target_normals = o3c.Tensor(
        [[0.15597, -0.0463812, -0.986672], [-0.213545, 0.887963, 0.407334],
         [0.423193, -0.121977, -0.897792], [0.202251, 0.27611, -0.939605],
         [0.275452, 0.207216, -0.938716], [0.326146, 0.0385317, -0.944534],
         [0.983129, -0.174668, -0.0543011], [0.898665, -0.0602029, 0.434485],
         [0.711325, 0.193223, -0.675783], [0.346158, 0.198724, -0.916888],
         [0.302085, 0.28938, -0.908297], [0.341044, 0.414138, -0.843907],
         [0.212191, 0.213068, -0.953717], [0.239759, 0.313187, -0.918929],
         [0.302290, 0.27265, -0.913391], [0.209796, 0.402747, -0.890944],
         [0.267025, 0.218226, -0.938656], [0.00126928, -0.976587, -0.21512],
         [0.321912, 0.194736, -0.926526], [0.831227, 0.236675, -0.503037],
         [0.987006, -0.155324, 0.0411639], [0.103384, -0.808796, -0.57893],
         [0.181245, 0.66226, -0.727023], [0.235471, 0.525053, -0.817846],
         [0.231954, 0.446165, -0.864369], [-0.261931, -0.725542, -0.636381],
         [0.120953, -0.864985, -0.487003], [0.858345, -0.227847, 0.459706],
         [-0.416259, -0.367408, -0.831709], [-0.476652, 0.206048, -0.854604],
         [-0.211959, -0.523378, -0.825317], [-0.964914, 0.0541031, -0.256931],
         [-0.0653566, -0.913961, -0.400504], [-0.846868, -0.170805, -0.503628],
         [0.0366971, 0.515834, -0.855902], [-0.0714554, -0.855019, -0.513651],
         [-0.0217377, -0.957744, -0.286799], [-0.0345231, -0.947096, -0.319088]
        ], dtype, device)

    source = o3d.t.geometry.PointCloud(device)
    target = o3d.t.geometry.PointCloud(device)

    source.point.positions = source_points
    target.point.positions = target_points
    target.point.normals = target_normals

    return source, target


@pytest.mark.parametrize("device", list_devices())
def test_icp_convergence_criteria_constructor(device):

    # Constructor.
    convergence_criteria = o3d.t.pipelines.registration.ICPConvergenceCriteria()

    # Checking default values.
    assert convergence_criteria.max_iteration == 30
    assert convergence_criteria.relative_fitness == 1e-06
    assert convergence_criteria.relative_rmse == 1e-06


@pytest.mark.parametrize("device", list_devices())
def test_registration_result_constructor(device):
    dtype = o3c.float64

    # Constructor.
    registration_result = o3d.t.pipelines.registration.RegistrationResult()

    # Checking default values.
    assert registration_result.inlier_rmse == 0.0
    assert registration_result.fitness == 0.0
    assert registration_result.transformation.allclose(
        o3c.Tensor.eye(4, dtype, o3c.Device("CPU:0")))


@pytest.mark.parametrize("device", list_devices())
def test_evaluate_registration(device):

    supported_dtypes = [o3c.float32, o3c.float64]
    for dtype in supported_dtypes:
        source_t, target_t = get_pcds(dtype, device)

        source_legacy = source_t.to_legacy()
        target_legacy = target_t.to_legacy()

        max_correspondence_distance = 3.0
        init_trans_legacy = np.eye(4)
        init_trans_t = o3c.Tensor.eye(4, o3c.float64, device)

        evaluation_t = o3d.t.pipelines.registration.evaluate_registration(
            source_t, target_t, max_correspondence_distance, init_trans_t)
        evaluation_legacy = o3d.pipelines.registration.evaluate_registration(
            source_legacy, target_legacy, max_correspondence_distance,
            init_trans_legacy)

        np.testing.assert_allclose(evaluation_t.inlier_rmse,
                                   evaluation_legacy.inlier_rmse, 0.001)
        np.testing.assert_allclose(evaluation_t.fitness,
                                   evaluation_legacy.fitness, 0.001)


@pytest.mark.parametrize("device", list_devices())
def test_icp_point_to_point(device):

    supported_dtypes = [o3c.float32, o3c.float64]
    for dtype in supported_dtypes:
        source_t, target_t = get_pcds(dtype, device)

        source_legacy = source_t.to_legacy()
        target_legacy = target_t.to_legacy()

        max_correspondence_distance = 3.0

        init_trans_legacy = np.array([[0.862, 0.011, -0.507, 0.5],
                                      [-0.139, 0.967, -0.215, 0.7],
                                      [0.487, 0.255, 0.835, -1.4],
                                      [0.0, 0.0, 0.0, 1.0]])
        init_trans_t = o3c.Tensor(init_trans_legacy,
                                  dtype=o3c.float64,
                                  device=device)

        reg_p2p_t = o3d.t.pipelines.registration.icp(
            source_t, target_t, max_correspondence_distance, init_trans_t,
            o3d.t.pipelines.registration.TransformationEstimationPointToPoint(),
            o3d.t.pipelines.registration.ICPConvergenceCriteria(
                max_iteration=2))

        reg_p2p_legacy = o3d.pipelines.registration.registration_icp(
            source_legacy, target_legacy, max_correspondence_distance,
            init_trans_legacy,
            o3d.pipelines.registration.TransformationEstimationPointToPoint(),
            o3d.pipelines.registration.ICPConvergenceCriteria(max_iteration=2))

        np.testing.assert_allclose(reg_p2p_t.inlier_rmse,
                                   reg_p2p_legacy.inlier_rmse, 0.001)
        np.testing.assert_allclose(reg_p2p_t.fitness, reg_p2p_legacy.fitness,
                                   0.001)


@pytest.mark.parametrize("device", list_devices())
def test_icp_point_to_plane(device):

    supported_dtypes = [o3c.float32, o3c.float64]
    for dtype in supported_dtypes:
        source_t, target_t = get_pcds(dtype, device)

        source_legacy = source_t.to_legacy()
        target_legacy = target_t.to_legacy()

        max_correspondence_distance = 3.0

        init_trans_legacy = np.array([[0.862, 0.011, -0.507, 0.5],
                                      [-0.139, 0.967, -0.215, 0.7],
                                      [0.487, 0.255, 0.835, -1.4],
                                      [0.0, 0.0, 0.0, 1.0]])
        init_trans_t = o3c.Tensor(init_trans_legacy,
                                  dtype=o3c.float64,
                                  device=device)

        reg_p2plane_t = o3d.t.pipelines.registration.icp(
            source_t, target_t, max_correspondence_distance, init_trans_t,
            o3d.t.pipelines.registration.TransformationEstimationPointToPlane(),
            o3d.t.pipelines.registration.ICPConvergenceCriteria(
                max_iteration=2))

        reg_p2plane_legacy = o3d.pipelines.registration.registration_icp(
            source_legacy, target_legacy, max_correspondence_distance,
            init_trans_legacy,
            o3d.pipelines.registration.TransformationEstimationPointToPlane(),
            o3d.pipelines.registration.ICPConvergenceCriteria(max_iteration=2))

        np.testing.assert_allclose(reg_p2plane_t.inlier_rmse,
                                   reg_p2plane_legacy.inlier_rmse, 0.001)
        np.testing.assert_allclose(reg_p2plane_t.fitness,
                                   reg_p2plane_legacy.fitness, 0.001)


@pytest.mark.parametrize("device", list_devices())
def test_get_information_matrix(device):

    supported_dtypes = [o3c.float32, o3c.float64]
    for dtype in supported_dtypes:
        source_t, target_t = get_pcds(dtype, device)

        source_legacy = source_t.to_legacy()
        target_legacy = target_t.to_legacy()

        max_correspondence_distance = 3.0

        transformation_legacy = np.array([[0.862, 0.011, -0.507, 0.5],
                                          [-0.139, 0.967, -0.215, 0.7],
                                          [0.487, 0.255, 0.835, -1.4],
                                          [0.0, 0.0, 0.0, 1.0]])
        transformation_t = o3c.Tensor(transformation_legacy,
                                      dtype=o3c.float64,
                                      device=device)

        info_matrix_t = o3d.t.pipelines.registration.get_information_matrix(
            source_t, target_t, max_correspondence_distance, transformation_t)

        info_matrix_legacy = o3d.pipelines.registration.get_information_matrix_from_point_clouds(
            source_legacy, target_legacy, max_correspondence_distance,
            transformation_legacy)

        np.testing.assert_allclose(info_matrix_t.cpu().numpy(),
                                   info_matrix_legacy, 1e-1, 1e-1)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import numpy as np
import os
import shutil
import sys
import zipfile
from os import listdir, makedirs
from os.path import exists, isfile, join, splitext, dirname, basename
import re
from warnings import warn
import json
import open3d as o3d
import copy

if (sys.version_info > (3, 0)):
    pyver = 3
    from urllib.request import Request, urlopen
else:
    pyver = 2
    from urllib2 import Request, urlopen


def edges_to_lineset(mesh, edges, color):
    ls = o3d.geometry.LineSet()
    ls.points = mesh.vertices
    ls.lines = edges
    ls.paint_uniform_color(color)
    return ls


def get_plane_mesh(height=0.2, width=1):
    mesh = o3d.geometry.TriangleMesh(
        vertices=o3d.utility.Vector3dVector(
            np.array(
                [[0, 0, 0], [0, height, 0], [width, height, 0], [width, 0, 0]],
                dtype=np.float32,
            )),
        triangles=o3d.utility.Vector3iVector(np.array([[0, 2, 1], [2, 0, 3]])),
    )
    mesh.compute_vertex_normals()
    return mesh


def get_non_manifold_edge_mesh():
    verts = np.array(
        [[-1, 0, 0], [0, 1, 0], [1, 0, 0], [0, -1, 0], [0, 0, 1]],
        dtype=np.float64,
    )
    triangles = np.array([[0, 1, 3], [1, 2, 3], [1, 3, 4]])
    mesh = o3d.geometry.TriangleMesh()
    mesh.vertices = o3d.utility.Vector3dVector(verts)
    mesh.triangles = o3d.utility.Vector3iVector(triangles)
    mesh.compute_vertex_normals()
    mesh.rotate(
        mesh.get_rotation_matrix_from_xyz((np.pi / 4, 0, np.pi / 4)),
        center=mesh.get_center(),
    )
    return mesh


def get_non_manifold_vertex_mesh():
    verts = np.array(
        [
            [-1, 0, -1],
            [1, 0, -1],
            [0, 1, -1],
            [0, 0, 0],
            [-1, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
        ],
        dtype=np.float64,
    )
    triangles = np.array([
        [0, 1, 2],
        [0, 1, 3],
        [1, 2, 3],
        [2, 0, 3],
        [4, 5, 6],
        [4, 5, 3],
        [5, 6, 3],
        [4, 6, 3],
    ])
    mesh = o3d.geometry.TriangleMesh()
    mesh.vertices = o3d.utility.Vector3dVector(verts)
    mesh.triangles = o3d.utility.Vector3iVector(triangles)
    mesh.compute_vertex_normals()
    mesh.rotate(
        mesh.get_rotation_matrix_from_xyz((np.pi / 4, 0, np.pi / 4)),
        center=mesh.get_center(),
    )
    return mesh


def get_open_box_mesh():
    mesh = o3d.geometry.TriangleMesh.create_box()
    mesh.triangles = o3d.utility.Vector3iVector(np.asarray(mesh.triangles)[:-2])
    mesh.compute_vertex_normals()
    mesh.rotate(
        mesh.get_rotation_matrix_from_xyz((0.8 * np.pi, 0, 0.66 * np.pi)),
        center=mesh.get_center(),
    )
    return mesh


def get_intersecting_boxes_mesh():
    mesh0 = o3d.geometry.TriangleMesh.create_box()
    T = np.eye(4)
    T[:, 3] += (0.5, 0.5, 0.5, 0)
    mesh1 = o3d.geometry.TriangleMesh.create_box()
    mesh1.transform(T)
    mesh = mesh0 + mesh1
    mesh.compute_vertex_normals()
    mesh.rotate(
        mesh.get_rotation_matrix_from_xyz((0.7 * np.pi, 0, 0.6 * np.pi)),
        center=mesh.get_center(),
    )
    return mesh


def file_downloader(url, out_dir="."):
    file_name = url.split('/')[-1]
    u = urlopen(url)
    f = open(os.path.join(out_dir, file_name), "wb")
    if pyver == 2:
        meta = u.info()
        file_size = int(meta.getheaders("Content-Length")[0])
    elif pyver == 3:
        file_size = int(u.getheader("Content-Length"))
    print("Downloading: %s " % file_name)

    file_size_dl = 0
    block_sz = 8192
    progress = 0
    while True:
        buffer = u.read(block_sz)
        if not buffer:
            break
        file_size_dl += len(buffer)
        f.write(buffer)
        if progress + 10 <= (file_size_dl * 100. / file_size):
            progress = progress + 10
            print(" %.1f / %.1f MB (%.0f %%)" % \
                    (file_size_dl/(1024*1024), file_size/(1024*1024), progress))
    f.close()


def unzip_data(path_zip, path_extract_to):
    print("Unzipping %s" % path_zip)
    zip_ref = zipfile.ZipFile(path_zip, 'r')
    zip_ref.extractall(path_extract_to)
    zip_ref.close()
    print("Extracted to %s" % path_extract_to)


def sorted_alphanum(file_list_ordered):
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(file_list_ordered, key=alphanum_key)


def get_file_list(path, extension=None):
    if extension is None:
        file_list = [path + f for f in listdir(path) if isfile(join(path, f))]
    else:
        file_list = [
            path + f
            for f in listdir(path)
            if isfile(join(path, f)) and splitext(f)[1] == extension
        ]
    file_list = sorted_alphanum(file_list)
    return file_list


def add_if_exists(path_dataset, folder_names):
    for folder_name in folder_names:
        if exists(join(path_dataset, folder_name)):
            path = join(path_dataset, folder_name)
            return path
    raise FileNotFoundError(
        f"None of the folders {folder_names} found in {path_dataset}")


def read_rgbd_image(color_file, depth_file, convert_rgb_to_intensity, config):
    color = o3d.io.read_image(color_file)
    depth = o3d.io.read_image(depth_file)
    rgbd_image = o3d.geometry.RGBDImage.create_from_color_and_depth(
        color,
        depth,
        depth_scale=config["depth_scale"],
        depth_trunc=config["depth_max"],
        convert_rgb_to_intensity=convert_rgb_to_intensity)
    return rgbd_image


def get_rgbd_folders(path_dataset):
    path_color = add_if_exists(path_dataset, ["image/", "rgb/", "color/"])
    path_depth = join(path_dataset, "depth/")
    return path_color, path_depth


def get_rgbd_file_lists(path_dataset):
    path_color, path_depth = get_rgbd_folders(path_dataset)
    color_files = get_file_list(path_color, ".jpg") + \
            get_file_list(path_color, ".png")
    depth_files = get_file_list(path_depth, ".png")
    return color_files, depth_files


def make_clean_folder(path_folder):
    if not exists(path_folder):
        makedirs(path_folder)
    else:
        shutil.rmtree(path_folder)
        makedirs(path_folder)


def check_folder_structure(path_dataset):
    if isfile(path_dataset) and path_dataset.endswith(".bag"):
        return
    path_color, path_depth = get_rgbd_folders(path_dataset)
    assert exists(path_depth), \
            "Path %s is not exist!" % path_depth
    assert exists(path_color), \
            "Path %s is not exist!" % path_color


def write_poses_to_log(filename, poses):
    with open(filename, 'w') as f:
        for i, pose in enumerate(poses):
            f.write('{} {} {}\n'.format(i, i, i + 1))
            f.write('{0:.8f} {1:.8f} {2:.8f} {3:.8f}\n'.format(
                pose[0, 0], pose[0, 1], pose[0, 2], pose[0, 3]))
            f.write('{0:.8f} {1:.8f} {2:.8f} {3:.8f}\n'.format(
                pose[1, 0], pose[1, 1], pose[1, 2], pose[1, 3]))
            f.write('{0:.8f} {1:.8f} {2:.8f} {3:.8f}\n'.format(
                pose[2, 0], pose[2, 1], pose[2, 2], pose[2, 3]))
            f.write('{0:.8f} {1:.8f} {2:.8f} {3:.8f}\n'.format(
                pose[3, 0], pose[3, 1], pose[3, 2], pose[3, 3]))


def read_poses_from_log(traj_log):
    import numpy as np

    trans_arr = []
    with open(traj_log) as f:
        content = f.readlines()

        # Load .log file.
        for i in range(0, len(content), 5):
            # format %d (src) %d (tgt) %f (fitness)
            data = list(map(float, content[i].strip().split(' ')))
            ids = (int(data[0]), int(data[1]))
            fitness = data[2]

            # format %f x 16
            T_gt = np.array(
                list(map(float, (''.join(
                    content[i + 1:i + 5])).strip().split()))).reshape((4, 4))

            trans_arr.append(T_gt)

    return trans_arr


flip_transform = [[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]


def draw_geometries_flip(pcds):
    pcds_transform = []
    for pcd in pcds:
        pcd_temp = copy.deepcopy(pcd)
        pcd_temp.transform(flip_transform)
        pcds_transform.append(pcd_temp)
    o3d.visualization.draw_geometries(pcds_transform)


def draw_registration_result(source, target, transformation):
    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)
    source_temp.paint_uniform_color([1, 0.706, 0])
    target_temp.paint_uniform_color([0, 0.651, 0.929])
    source_temp.transform(transformation)
    source_temp.transform(flip_transform)
    target_temp.transform(flip_transform)
    o3d.visualization.draw_geometries([source_temp, target_temp])


def draw_registration_result_original_color(source, target, transformation):
    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)
    source_temp.transform(transformation)
    source_temp.transform(flip_transform)
    target_temp.transform(flip_transform)
    o3d.visualization.draw_geometries([source_temp, target_temp])


class CameraPose:

    def __init__(self, meta, mat):
        self.metadata = meta
        self.pose = mat

    def __str__(self):
        return 'Metadata : ' + ' '.join(map(str, self.metadata)) + '\n' + \
            "Pose : " + "\n" + np.array_str(self.pose)


def read_trajectory(filename):
    traj = []
    with open(filename, 'r') as f:
        metastr = f.readline()
        while metastr:
            metadata = list(map(int, metastr.split()))
            mat = np.zeros(shape=(4, 4))
            for i in range(4):
                matstr = f.readline()
                mat[i, :] = np.fromstring(matstr, dtype=float, sep=' \t')
            traj.append(CameraPose(metadata, mat))
            metastr = f.readline()
    return traj


def write_trajectory(traj, filename):
    with open(filename, 'w') as f:
        for x in traj:
            p = x.pose.tolist()
            f.write(' '.join(map(str, x.metadata)) + '\n')
            f.write('\n'.join(
                ' '.join(map('{0:.12f}'.format, p[i])) for i in range(4)))
            f.write('\n')


def initialize_opencv():
    opencv_installed = True
    try:
        import cv2
    except ImportError:
        pass
        print("OpenCV is not detected. Using Identity as an initial")
        opencv_installed = False
    if opencv_installed:
        print("OpenCV is detected. Using ORB + 5pt algorithm")
    return opencv_installed



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

# Workaround when multiple copies of the OpenMP runtime have been linked to
# the program, which happens when PyTorch loads OpenMP runtime first. Not that
# this method is "unsafe, unsupported, undocumented", but we found it to be
# generally safe to use. This should be deprecated once we found a way to
# "ensure that only a single OpenMP runtime is linked into the process".
#
# https://github.com/llvm-mirror/openmp/blob/8453ca8594e1a5dd8a250c39bf8fcfbfb1760e60/runtime/src/i18n/en_US.txt#L449
# https://github.com/dmlc/xgboost/issues/1715
import os
import sys
import re

os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
# Enable thread composability manager to coordinate Intel OpenMP and TBB threads. Only works with Intel OpenMP.
# TBB must not be already loaded.
os.environ["TCM_ENABLE"] = "1"
from ctypes import CDLL
from ctypes.util import find_library
from pathlib import Path
import warnings
from open3d._build_config import _build_config

if sys.platform == "win32":  # Unix: Use rpath to find libraries
    _win32_dll_dir = os.add_dll_directory(str(Path(__file__).parent))

__DEVICE_API__ = "cpu"
if _build_config["BUILD_CUDA_MODULE"]:
    # Load CPU pybind dll gracefully without introducing new python variable.
    # Do this before loading the CUDA pybind dll to correctly resolve symbols
    try:  # StopIteration if cpu version not available
        CDLL(str(next((Path(__file__).parent / "cpu").glob("pybind*"))))
    except StopIteration:
        warnings.warn(
            "Open3D was built with CUDA support, but Open3D CPU Python "
            "bindings were not found. Open3D will not work on systems without"
            " CUDA devices.",
            ImportWarning,
        )
    try:
        if sys.platform == "win32" and sys.version_info >= (3, 8):
            # Since Python 3.8, the PATH environment variable is not used to find DLLs anymore.
            # To allow Windows users to use Open3D with CUDA without running into dependency-problems,
            # look for the CUDA bin directory in PATH and explicitly add it to the DLL search path.
            cuda_bin_path = None
            for path in os.environ['PATH'].split(';'):
                # search heuristic: look for a path containing "cuda" and "bin" in this order.
                if re.search(r'cuda.*bin', path, re.IGNORECASE):
                    cuda_bin_path = path
                    break

            if cuda_bin_path:
                os.add_dll_directory(cuda_bin_path)

        # Check CUDA availability without importing CUDA pybind symbols to
        # prevent "symbol already registered" errors if first import fails.
        _pybind_cuda = CDLL(
            str(next((Path(__file__).parent / "cuda").glob("pybind*"))))
        if _pybind_cuda.open3d_core_cuda_device_count() > 0:
            from open3d.cuda.pybind import (
                core,
                camera,
                data,
                geometry,
                io,
                pipelines,
                utility,
                t,
            )
            from open3d.cuda import pybind

            __DEVICE_API__ = "cuda"
        else:
            warnings.warn(
                "Open3D was built with CUDA support, but no suitable CUDA "
                "devices found. If your system has CUDA devices, check your "
                "CUDA drivers and runtime.",
                ImportWarning,
            )
    except OSError as os_error:
        warnings.warn(
            f"Open3D was built with CUDA support, but an error ocurred while loading the Open3D CUDA Python bindings. This is usually because the CUDA libraries could not be found. Check your CUDA installation. Falling back to the CPU pybind library. Reported error: {os_error}.",
            ImportWarning,
        )
    except StopIteration:
        warnings.warn(
            "Open3D was built with CUDA support, but Open3D CUDA Python "
            "binding library not found! Falling back to the CPU Python "
            "binding library.",
            ImportWarning,
        )

if __DEVICE_API__ == "cpu":
    from open3d.cpu.pybind import (
        core,
        camera,
        data,
        geometry,
        io,
        pipelines,
        utility,
        t,
    )
    from open3d.cpu import pybind


def _insert_pybind_names(skip_names=()):
    """Introduce pybind names as open3d names. Skip names corresponding to
    python subpackages, since they have a different import mechanism."""
    submodules = {}
    for modname in sys.modules:
        if "open3d." + __DEVICE_API__ + ".pybind" in modname:
            if any("." + skip_name in modname for skip_name in skip_names):
                continue
            subname = modname.replace(__DEVICE_API__ + ".pybind.", "")
            if subname not in sys.modules:
                submodules[subname] = sys.modules[modname]
    sys.modules.update(submodules)


import open3d.visualization

_insert_pybind_names(skip_names=("ml",))

__version__ = "@PROJECT_VERSION@"

if int(sys.version_info[0]) < 3:
    raise Exception("Open3D only supports Python 3.")

if (_build_config["BUILD_JUPYTER_EXTENSION"] and os.environ.get(
        "OPEN3D_DISABLE_WEB_VISUALIZER", "False").lower() != "true"):
    import platform

    if not (platform.machine().startswith("arm") or
            platform.machine().startswith("aarch")):
        try:
            shell = get_ipython().__class__.__name__
            if shell == "ZMQInteractiveShell":
                print("Jupyter environment detected. "
                      "Enabling Open3D WebVisualizer.")
                # Set default window system.
                open3d.visualization.webrtc_server.enable_webrtc()
                # HTTP handshake server is needed when Open3D is serving the
                # visualizer webpage. Disable since Jupyter is serving.
                open3d.visualization.webrtc_server.disable_http_handshake()
        except NameError:
            pass
    else:
        warnings.warn("Open3D WebVisualizer is not supported on ARM for now.",
                      RuntimeWarning)

# OPEN3D_ML_ROOT points to the root of the Open3D-ML repo.
# If set this will override the integrated Open3D-ML.
if "OPEN3D_ML_ROOT" in os.environ:
    print("Using external Open3D-ML in {}".format(os.environ["OPEN3D_ML_ROOT"]))
    sys.path.append(os.environ["OPEN3D_ML_ROOT"])
import open3d.ml

# Finally insert pybind names corresponding to ml
_insert_pybind_names()


def _jupyter_labextension_paths():
    """Called by Jupyter Lab Server to detect if it is a valid labextension and
    to install the widget.

    Returns:
        src: Source directory name to copy files from. Webpack outputs generated
            files into this directory and Jupyter Lab copies from this directory
            during widget installation.
        dest: Destination directory name to install widget files to. Jupyter Lab
            copies from `src` directory into <jupyter path>/labextensions/<dest>
            directory during widget installation.
    """
    return [{
        "src": "labextension",
        "dest": "open3d",
    }]


def _jupyter_nbextension_paths():
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension
    and to install the widget.

    Returns:
        section: The section of the Jupyter Notebook Server to change.
            Must be "notebook" for widget extensions.
        src: Source directory name to copy files from. Webpack outputs generated
            files into this directory and Jupyter Notebook copies from this
            directory during widget installation.
        dest: Destination directory name to install widget files to. Jupyter
            Notebook copies from `src` directory into
            <jupyter path>/nbextensions/<dest> directory during widget
            installation.
        require: Path to importable AMD Javascript module inside the
            <jupyter path>/nbextensions/<dest> directory.
    """
    return [{
        "section": "notebook",
        "src": "nbextension",
        "dest": "open3d",
        "require": "open3d/extension",
    }]


if sys.platform == "win32":
    _win32_dll_dir.close()
del os, sys, CDLL, find_library, Path, warnings, _insert_pybind_names



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

# examples/python/t_reconstruction_system/dense_slam.py

# P.S. This example is used in documentation, so, please ensure the changes are
# synchronized.

import os
import numpy as np
import open3d as o3d
import time

from config import ConfigParser
from common import (get_default_dataset, load_rgbd_file_names, save_poses,
                    load_intrinsic, extract_trianglemesh, extract_rgbd_frames)


def slam(depth_file_names, color_file_names, intrinsic, config):
    n_files = len(color_file_names)
    device = o3d.core.Device(config.device)

    T_frame_to_model = o3d.core.Tensor(np.identity(4))
    model = o3d.t.pipelines.slam.Model(config.voxel_size, 16,
                                       config.block_count, T_frame_to_model,
                                       device)
    depth_ref = o3d.t.io.read_image(depth_file_names[0])
    input_frame = o3d.t.pipelines.slam.Frame(depth_ref.rows, depth_ref.columns,
                                             intrinsic, device)
    raycast_frame = o3d.t.pipelines.slam.Frame(depth_ref.rows,
                                               depth_ref.columns, intrinsic,
                                               device)

    poses = []

    for i in range(n_files):
        start = time.time()

        depth = o3d.t.io.read_image(depth_file_names[i]).to(device)
        color = o3d.t.io.read_image(color_file_names[i]).to(device)

        input_frame.set_data_from_image('depth', depth)
        input_frame.set_data_from_image('color', color)

        if i > 0:
            result = model.track_frame_to_model(input_frame, raycast_frame,
                                                config.depth_scale,
                                                config.depth_max,
                                                config.odometry_distance_thr)
            T_frame_to_model = T_frame_to_model @ result.transformation

        poses.append(T_frame_to_model.cpu().numpy())
        model.update_frame_pose(i, T_frame_to_model)
        model.integrate(input_frame, config.depth_scale, config.depth_max,
                        config.trunc_voxel_multiplier)
        model.synthesize_model_frame(raycast_frame, config.depth_scale,
                                     config.depth_min, config.depth_max,
                                     config.trunc_voxel_multiplier, False)
        stop = time.time()
        print('{:04d}/{:04d} slam takes {:.4}s'.format(i, n_files,
                                                       stop - start))

    return model.voxel_grid, poses


if __name__ == '__main__':
    parser = ConfigParser()
    parser.add(
        '--config',
        is_config_file=True,
        help='YAML config file path. Please refer to default_config.yml as a '
        'reference. It overrides the default config file, but will be '
        'overridden by other command line inputs.')
    parser.add('--default_dataset',
               help='Default dataset is used when config file is not provided. '
               'Default dataset may be selected from the following options: '
               '[lounge, bedroom, jack_jack]',
               default='lounge')
    parser.add('--path_npz',
               help='path to the npz file that stores voxel block grid.',
               default='output.npz')
    config = parser.get_config()

    if config.path_dataset == '':
        config = get_default_dataset(config)

    # Extract RGB-D frames and intrinsic from bag file.
    if config.path_dataset.endswith(".bag"):
        assert os.path.isfile(
            config.path_dataset), f"File {config.path_dataset} not found."
        print("Extracting frames from RGBD video file")
        config.path_dataset, config.path_intrinsic, config.depth_scale = extract_rgbd_frames(
            config.path_dataset)

    depth_file_names, color_file_names = load_rgbd_file_names(config)
    intrinsic = load_intrinsic(config)

    if not os.path.exists(config.path_npz):
        volume, poses = slam(depth_file_names, color_file_names, intrinsic,
                             config)
        print('Saving to {}...'.format(config.path_npz))
        volume.save(config.path_npz)
        save_poses('output.log', poses)
        print('Saving finished')
    else:
        volume = o3d.t.geometry.VoxelBlockGrid.load(config.path_npz)

    mesh = extract_trianglemesh(volume, config, 'output.ply')
    o3d.visualization.draw([mesh])



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import open3d.core as o3c
import numpy as np
import pytest
import operator

import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/..")
from open3d_benchmark import list_tensor_sizes, list_non_bool_dtypes, to_numpy_dtype


class BinaryEWOps:

    @staticmethod
    def logical_and(lhs, rhs):
        return lhs.logical_and(rhs)

    @staticmethod
    def logical_or(lhs, rhs):
        return lhs.logical_or(rhs)

    @staticmethod
    def logical_xor(lhs, rhs):
        return lhs.logical_xor(rhs)


def list_binary_ops():
    return [
        operator.add,
        operator.sub,
        operator.mul,
        operator.truediv,
        BinaryEWOps.logical_and,
        BinaryEWOps.logical_or,
        BinaryEWOps.logical_xor,
        operator.gt,
        operator.lt,
        operator.ge,
        operator.le,
        operator.eq,
        operator.ne,
    ]


def to_numpy_binary_op(op):
    conversions = {
        operator.add: operator.add,
        operator.sub: operator.sub,
        operator.mul: operator.mul,
        operator.truediv: operator.truediv,
        BinaryEWOps.logical_and: np.logical_and,
        BinaryEWOps.logical_or: np.logical_or,
        BinaryEWOps.logical_xor: np.logical_xor,
        operator.gt: operator.gt,
        operator.lt: operator.lt,
        operator.ge: operator.ge,
        operator.le: operator.le,
        operator.eq: operator.eq,
        operator.ne: operator.ne,
    }
    return conversions[op]


@pytest.mark.parametrize("size", list_tensor_sizes())
@pytest.mark.parametrize("dtype", list_non_bool_dtypes())
@pytest.mark.parametrize("op", list_binary_ops())
def test_binary_ew_ops(benchmark, size, dtype, op):
    np_a = np.array(np.random.uniform(1, 127, size),
                    dtype=to_numpy_dtype(dtype))
    np_b = np.array(np.random.uniform(1, 127, size),
                    dtype=to_numpy_dtype(dtype))
    a = o3c.Tensor(np_a, dtype=dtype, device=o3c.Device("CPU:0"))
    b = o3c.Tensor(np_b, dtype=dtype, device=o3c.Device("CPU:0"))
    benchmark(op, a, b)


@pytest.mark.parametrize("size", list_tensor_sizes())
@pytest.mark.parametrize("dtype", list_non_bool_dtypes())
@pytest.mark.parametrize("op", list_binary_ops())
def test_binary_ew_ops_numpy(benchmark, size, dtype, op):
    np_a = np.array(np.random.uniform(1, 127, size),
                    dtype=to_numpy_dtype(dtype))
    np_b = np.array(np.random.uniform(1, 127, size),
                    dtype=to_numpy_dtype(dtype))
    benchmark(to_numpy_binary_op(op), np_a, np_b)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt

if __name__ == "__main__":
    # Create meshes and convert to open3d.t.geometry.TriangleMesh .
    cube = o3d.geometry.TriangleMesh.create_box().translate([0, 0, 0])
    cube = o3d.t.geometry.TriangleMesh.from_legacy(cube)
    torus = o3d.geometry.TriangleMesh.create_torus().translate([0, 0, 2])
    torus = o3d.t.geometry.TriangleMesh.from_legacy(torus)
    sphere = o3d.geometry.TriangleMesh.create_sphere(radius=0.5).translate(
        [1, 2, 3])
    sphere = o3d.t.geometry.TriangleMesh.from_legacy(sphere)

    scene = o3d.t.geometry.RaycastingScene()
    scene.add_triangles(cube)
    scene.add_triangles(torus)
    _ = scene.add_triangles(sphere)

    rays = o3d.t.geometry.RaycastingScene.create_rays_pinhole(
        fov_deg=90,
        center=[0, 0, 2],
        eye=[2, 3, 0],
        up=[0, 1, 0],
        width_px=640,
        height_px=480,
    )
    # We can directly pass the rays tensor to the cast_rays function.
    ans = scene.cast_rays(rays)
    plt.imshow(ans['t_hit'].numpy())
    plt.show()
    plt.imshow(np.abs(ans['primitive_normals'].numpy()))
    plt.show()
    plt.imshow(np.abs(ans['geometry_ids'].numpy()), vmax=3)
    plt.show()



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import numpy as np
import os
import sys

pyexample_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(pyexample_path)

import open3d_example as o3dex


def check_properties(name, mesh):
    mesh.compute_vertex_normals()

    edge_manifold = mesh.is_edge_manifold(allow_boundary_edges=True)
    edge_manifold_boundary = mesh.is_edge_manifold(allow_boundary_edges=False)
    vertex_manifold = mesh.is_vertex_manifold()
    self_intersecting = mesh.is_self_intersecting()
    watertight = mesh.is_watertight()
    orientable = mesh.is_orientable()

    print(name)
    print(f"  edge_manifold:          {edge_manifold}")
    print(f"  edge_manifold_boundary: {edge_manifold_boundary}")
    print(f"  vertex_manifold:        {vertex_manifold}")
    print(f"  self_intersecting:      {self_intersecting}")
    print(f"  watertight:             {watertight}")
    print(f"  orientable:             {orientable}")

    geoms = [mesh]
    if not edge_manifold:
        edges = mesh.get_non_manifold_edges(allow_boundary_edges=True)
        geoms.append(o3dex.edges_to_lineset(mesh, edges, (1, 0, 0)))
    if not edge_manifold_boundary:
        edges = mesh.get_non_manifold_edges(allow_boundary_edges=False)
        geoms.append(o3dex.edges_to_lineset(mesh, edges, (0, 1, 0)))
    if not vertex_manifold:
        verts = np.asarray(mesh.get_non_manifold_vertices())
        pcl = o3d.geometry.PointCloud(
            points=o3d.utility.Vector3dVector(np.asarray(mesh.vertices)[verts]))
        pcl.paint_uniform_color((0, 0, 1))
        geoms.append(pcl)
    if self_intersecting:
        intersecting_triangles = np.asarray(
            mesh.get_self_intersecting_triangles())
        intersecting_triangles = intersecting_triangles[0:1]
        intersecting_triangles = np.unique(intersecting_triangles)
        print("  # visualize self-intersecting triangles")
        triangles = np.asarray(mesh.triangles)[intersecting_triangles]
        edges = [
            np.vstack((triangles[:, i], triangles[:, j]))
            for i, j in [(0, 1), (1, 2), (2, 0)]
        ]
        edges = np.hstack(edges).T
        edges = o3d.utility.Vector2iVector(edges)
        geoms.append(o3dex.edges_to_lineset(mesh, edges, (1, 0, 1)))
    o3d.visualization.draw_geometries(geoms, mesh_show_back_face=True)


if __name__ == "__main__":
    knot_mesh = o3d.data.KnotMesh()
    mesh = o3d.io.read_triangle_mesh(knot_mesh.path)
    check_properties('KnotMesh', mesh)
    check_properties('Mobius',
                     o3d.geometry.TriangleMesh.create_mobius(twists=1))
    check_properties("non-manifold edge", o3dex.get_non_manifold_edge_mesh())
    check_properties("non-manifold vertex",
                     o3dex.get_non_manifold_vertex_mesh())
    check_properties("open box", o3dex.get_open_box_mesh())
    check_properties("intersecting_boxes", o3dex.get_intersecting_boxes_mesh())



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d

if __name__ == "__main__":
    bunny = o3d.data.BunnyMesh()
    mesh = o3d.io.read_triangle_mesh(bunny.path)
    mesh.compute_vertex_normals()

    print("Displaying input mesh ...")
    o3d.visualization.draw([mesh])

    print("Displaying pointcloud using uniform sampling ...")
    pcd = mesh.sample_points_uniformly(number_of_points=1000)
    o3d.visualization.draw([pcd], point_size=5)

    print("Displaying pointcloud using Poisson disk sampling ...")
    pcd = mesh.sample_points_poisson_disk(number_of_points=1000, init_factor=5)
    o3d.visualization.draw([pcd], point_size=5)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------

import open3d as o3d
import numpy as np
import pytest

import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/..")
from open3d_test import list_devices


def test_size_vector():
    # List
    sv = o3d.core.SizeVector([-1, 2, 3])
    assert "{}".format(sv) == "SizeVector[-1, 2, 3]"

    # Tuple
    sv = o3d.core.SizeVector((-1, 2, 3))
    assert "{}".format(sv) == "SizeVector[-1, 2, 3]"

    # Numpy 1D array
    sv = o3d.core.SizeVector(np.array([-1, 2, 3]))
    assert "{}".format(sv) == "SizeVector[-1, 2, 3]"

    # Empty
    sv = o3d.core.SizeVector()
    assert "{}".format(sv) == "SizeVector[]"
    sv = o3d.core.SizeVector([])
    assert "{}".format(sv) == "SizeVector[]"
    sv = o3d.core.SizeVector(())
    assert "{}".format(sv) == "SizeVector[]"
    sv = o3d.core.SizeVector(np.array([]))
    assert "{}".format(sv) == "SizeVector[]"

    # 1-dimensional SizeVector
    assert o3d.core.SizeVector(3) == (3,)
    assert o3d.core.SizeVector((3)) == (3,)
    assert o3d.core.SizeVector((3,)) == (3,)
    assert o3d.core.SizeVector([3]) == (3,)
    assert o3d.core.SizeVector([
        3,
    ]) == (3,)

    # Not integer: thorws exception
    with pytest.raises(Exception):
        sv = o3d.core.SizeVector([1.9, 2, 3])

    with pytest.raises(Exception):
        sv = o3d.core.SizeVector([-1.5, 2, 3])

    # 2D list exception
    with pytest.raises(Exception):
        sv = o3d.core.SizeVector([[1, 2], [3, 4]])

    # 2D Numpy array exception
    with pytest.raises(Exception):
        sv = o3d.core.SizeVector(np.array([[1, 2], [3, 4]]))

    # Garbage input
    with pytest.raises(Exception):
        sv = o3d.core.SizeVector(["foo", "bar"])


@pytest.mark.parametrize("device", list_devices(enable_sycl=True))
def test_implicit_conversion(device):
    # Reshape
    t = o3d.core.Tensor.ones((3, 4), device=device)
    assert t.reshape(o3d.core.SizeVector((4, 3))).shape == (4, 3)
    assert t.reshape(o3d.core.SizeVector([4, 3])).shape == (4, 3)
    assert t.reshape((4, 3)).shape == (4, 3)
    assert t.reshape([4, 3]).shape == (4, 3)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.reshape((4, 3.0))
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.reshape((4.0, 3.0))
    with pytest.raises(RuntimeError, match="Invalid shape dimension"):
        t.reshape((4, -3))

    # 0-dimensional
    assert o3d.core.Tensor.ones((), device=device).shape == ()
    assert o3d.core.Tensor.ones([], device=device).shape == ()

    # 1-dimensional
    assert o3d.core.Tensor.ones(3, device=device).shape == (3,)
    assert o3d.core.Tensor.ones((3), device=device).shape == (3,)
    assert o3d.core.Tensor.ones((3,), device=device).shape == (3,)
    assert o3d.core.Tensor.ones([3], device=device).shape == (3,)
    assert o3d.core.Tensor.ones([
        3,
    ], device=device).shape == (3,)

    # Tensor creation
    assert o3d.core.Tensor.empty((3, 4), device=device).shape == (3, 4)
    assert o3d.core.Tensor.ones((3, 4), device=device).shape == (3, 4)
    assert o3d.core.Tensor.zeros((3, 4), device=device).shape == (3, 4)
    assert o3d.core.Tensor.full((3, 4), 10, device=device).shape == (3, 4)

    # Reduction
    t = o3d.core.Tensor.ones((3, 4, 5), device=device)
    assert t.sum(o3d.core.SizeVector([0, 2])).shape == (4,)
    assert t.sum(o3d.core.SizeVector([0, 2]), keepdim=True).shape == (1, 4, 1)
    assert t.sum((0, 2)).shape == (4,)
    assert t.sum([0, 2]).shape == (4,)
    assert t.sum((0, 2), keepdim=True).shape == (1, 4, 1)
    assert t.sum([0, 2], keepdim=True).shape == (1, 4, 1)



# ----------------------------------------------------------------------------
# -                        Open3D: www.open3d.org                            -
# ----------------------------------------------------------------------------
# Copyright (c) 2018-2024 www.open3d.org
# SPDX-License-Identifier: MIT
# ----------------------------------------------------------------------------
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: plugin_data.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x11plugin_data.proto\x12\x12tensorboard.open3d\"\xb2\r\n\x10Open3DPluginData\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12S\n\x13property_references\x18\x02 \x03(\x0b\x32\x36.tensorboard.open3d.Open3DPluginData.PropertyReference\x12\x44\n\x0b\x62\x61tch_index\x18\x03 \x01(\x0b\x32/.tensorboard.open3d.Open3DPluginData.BatchIndex\x1aw\n\x11PropertyReference\x12P\n\x11geometry_property\x18\x01 \x01(\x0e\x32\x35.tensorboard.open3d.Open3DPluginData.GeometryProperty\x12\x10\n\x08step_ref\x18\x02 \x01(\x04\x1a\x7f\n\tStartSize\x12\r\n\x05start\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\x12\x15\n\rmasked_crc32c\x18\x03 \x01(\r\x12\x11\n\taux_start\x18\x04 \x01(\x04\x12\x10\n\x08\x61ux_size\x18\x05 \x01(\x04\x12\x19\n\x11\x61ux_masked_crc32c\x18\x06 \x01(\r\x1a\x62\n\nBatchIndex\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x42\n\nstart_size\x18\x02 \x03(\x0b\x32..tensorboard.open3d.Open3DPluginData.StartSize\"\x93\t\n\x10GeometryProperty\x12\x14\n\x10vertex_positions\x10\x00\x12\x12\n\x0evertex_normals\x10\x01\x12\x11\n\rvertex_colors\x10\x02\x12\x16\n\x12vertex_texture_uvs\x10\x03\x12\x14\n\x10triangle_indices\x10\x04\x12\x13\n\x0ftriangle_colors\x10\x05\x12\x14\n\x10triangle_normals\x10\x06\x12\x18\n\x14triangle_texture_uvs\x10\x07\x12\x10\n\x0cline_indices\x10\x08\x12\x0f\n\x0bline_colors\x10\t\x12\x1c\n\x18material_scalar_metallic\x10@\x12\x1d\n\x19material_scalar_roughness\x10\x41\x12\x1f\n\x1bmaterial_scalar_reflectance\x10\x42\x12\x1e\n\x1amaterial_scalar_clear_coat\x10\x44\x12(\n$material_scalar_clear_coat_roughness\x10\x45\x12\x1e\n\x1amaterial_scalar_anisotropy\x10\x46\x12%\n!material_scalar_ambient_occlusion\x10G\x12 \n\x1cmaterial_scalar_transmission\x10I\x12\x1d\n\x19material_scalar_thickness\x10K\x12\'\n#material_scalar_absorption_distance\x10L\x12\x1e\n\x1amaterial_vector_base_color\x10P\x12\x1a\n\x16material_vector_normal\x10S\x12$\n material_vector_absorption_color\x10X\x12!\n\x1dmaterial_texture_map_metallic\x10`\x12\"\n\x1ematerial_texture_map_roughness\x10\x61\x12$\n material_texture_map_reflectance\x10\x62\x12#\n\x1fmaterial_texture_map_clear_coat\x10\x64\x12-\n)material_texture_map_clear_coat_roughness\x10\x65\x12#\n\x1fmaterial_texture_map_anisotropy\x10\x66\x12*\n&material_texture_map_ambient_occlusion\x10g\x12%\n!material_texture_map_transmission\x10i\x12\"\n\x1ematerial_texture_map_thickness\x10k\x12\x1f\n\x1bmaterial_texture_map_albedo\x10l\x12\x1f\n\x1bmaterial_texture_map_normal\x10o\x12)\n%material_texture_map_absorption_color\x10t\x12\'\n#material_texture_map_ao_rough_metal\x10u\"\x92\x01\n\rInferenceData\x12K\n\x10inference_result\x18\x01 \x03(\x0b\x32\x31.tensorboard.open3d.InferenceData.InferenceResult\x1a\x34\n\x0fInferenceResult\x12\r\n\x05label\x18\x01 \x01(\x05\x12\x12\n\nconfidence\x18\x02 \x01(\x02\"\x8f\x01\n\x0cLabelToNames\x12J\n\x0elabel_to_names\x18\x01 \x03(\x0b\x32\x32.tensorboard.open3d.LabelToNames.LabelToNamesEntry\x1a\x33\n\x11LabelToNamesEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3'
)

_OPEN3DPLUGINDATA = DESCRIPTOR.message_types_by_name['Open3DPluginData']
_OPEN3DPLUGINDATA_PROPERTYREFERENCE = _OPEN3DPLUGINDATA.nested_types_by_name[
    'PropertyReference']
_OPEN3DPLUGINDATA_STARTSIZE = _OPEN3DPLUGINDATA.nested_types_by_name[
    'StartSize']
_OPEN3DPLUGINDATA_BATCHINDEX = _OPEN3DPLUGINDATA.nested_types_by_name[
    'BatchIndex']
_INFERENCEDATA = DESCRIPTOR.message_types_by_name['InferenceData']
_INFERENCEDATA_INFERENCERESULT = _INFERENCEDATA.nested_types_by_name[
    'InferenceResult']
_LABELTONAMES = DESCRIPTOR.message_types_by_name['LabelToNames']
_LABELTONAMES_LABELTONAMESENTRY = _LABELTONAMES.nested_types_by_name[
    'LabelToNamesEntry']
_OPEN3DPLUGINDATA_GEOMETRYPROPERTY = _OPEN3DPLUGINDATA.enum_types_by_name[
    'GeometryProperty']
Open3DPluginData = _reflection.GeneratedProtocolMessageType(
    'Open3DPluginData',
    (_message.Message,),
    {
        'PropertyReference':
            _reflection.GeneratedProtocolMessageType(
                'PropertyReference',
                (_message.Message,),
                {
                    'DESCRIPTOR': _OPEN3DPLUGINDATA_PROPERTYREFERENCE,
                    '__module__': 'plugin_data_pb2'
                    # @@protoc_insertion_point(class_scope:tensorboard.open3d.Open3DPluginData.PropertyReference)
                }),
        'StartSize':
            _reflection.GeneratedProtocolMessageType(
                'StartSize',
                (_message.Message,),
                {
                    'DESCRIPTOR': _OPEN3DPLUGINDATA_STARTSIZE,
                    '__module__': 'plugin_data_pb2'
                    # @@protoc_insertion_point(class_scope:tensorboard.open3d.Open3DPluginData.StartSize)
                }),
        'BatchIndex':
            _reflection.GeneratedProtocolMessageType(
                'BatchIndex',
                (_message.Message,),
                {
                    'DESCRIPTOR': _OPEN3DPLUGINDATA_BATCHINDEX,
                    '__module__': 'plugin_data_pb2'
                    # @@protoc_insertion_point(class_scope:tensorboard.open3d.Open3DPluginData.BatchIndex)
                }),
        'DESCRIPTOR':
            _OPEN3DPLUGINDATA,
        '__module__':
            'plugin_data_pb2'
        # @@protoc_insertion_point(class_scope:tensorboard.open3d.Open3DPluginData)
    })
_sym_db.RegisterMessage(Open3DPluginData)
_sym_db.RegisterMessage(Open3DPluginData.PropertyReference)
_sym_db.RegisterMessage(Open3DPluginData.StartSize)
_sym_db.RegisterMessage(Open3DPluginData.BatchIndex)

InferenceData = _reflection.GeneratedProtocolMessageType(
    'InferenceData',
    (_message.Message,),
    {
        'InferenceResult':
            _reflection.GeneratedProtocolMessageType(
                'InferenceResult',
                (_message.Message,),
                {
                    'DESCRIPTOR': _INFERENCEDATA_INFERENCERESULT,
                    '__module__': 'plugin_data_pb2'
                    # @@protoc_insertion_point(class_scope:tensorboard.open3d.InferenceData.InferenceResult)
                }),
        'DESCRIPTOR':
            _INFERENCEDATA,
        '__module__':
            'plugin_data_pb2'
        # @@protoc_insertion_point(class_scope:tensorboard.open3d.InferenceData)
    })
_sym_db.RegisterMessage(InferenceData)
_sym_db.RegisterMessage(InferenceData.InferenceResult)

LabelToNames = _reflection.GeneratedProtocolMessageType(
    'LabelToNames',
    (_message.Message,),
    {
        'LabelToNamesEntry':
            _reflection.GeneratedProtocolMessageType(
                'LabelToNamesEntry',
                (_message.Message,),
                {
                    'DESCRIPTOR': _LABELTONAMES_LABELTONAMESENTRY,
                    '__module__': 'plugin_data_pb2'
                    # @@protoc_insertion_point(class_scope:tensorboard.open3d.LabelToNames.LabelToNamesEntry)
                }),
        'DESCRIPTOR':
            _LABELTONAMES,
        '__module__':
            'plugin_data_pb2'
        # @@protoc_insertion_point(class_scope:tensorboard.open3d.LabelToNames)
    })
_sym_db.RegisterMessage(LabelToNames)
_sym_db.RegisterMessage(LabelToNames.LabelToNamesEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _LABELTONAMES_LABELTONAMESENTRY._options = None
    _LABELTONAMES_LABELTONAMESENTRY._serialized_options = b'8\001'
    _OPEN3DPLUGINDATA._serialized_start = 42
    _OPEN3DPLUGINDATA._serialized_end = 1756
    _OPEN3DPLUGINDATA_PROPERTYREFERENCE._serialized_start = 234
    _OPEN3DPLUGINDATA_PROPERTYREFERENCE._serialized_end = 353
    _OPEN3DPLUGINDATA_STARTSIZE._serialized_start = 355
    _OPEN3DPLUGINDATA_STARTSIZE._serialized_end = 482
    _OPEN3DPLUGINDATA_BATCHINDEX._serialized_start = 484
    _OPEN3DPLUGINDATA_BATCHINDEX._serialized_end = 582
    _OPEN3DPLUGINDATA_GEOMETRYPROPERTY._serialized_start = 585
    _OPEN3DPLUGINDATA_GEOMETRYPROPERTY._serialized_end = 1756
    _INFERENCEDATA._serialized_start = 1759
    _INFERENCEDATA._serialized_end = 1905
    _INFERENCEDATA_INFERENCERESULT._serialized_start = 1853
    _INFERENCEDATA_INFERENCERESULT._serialized_end = 1905
    _LABELTONAMES._serialized_start = 1908
    _LABELTONAMES._serialized_end = 2051
    _LABELTONAMES_LABELTONAMESENTRY._serialized_start = 2000
    _LABELTONAMES_LABELTONAMESENTRY._serialized_end = 2051
# @@protoc_insertion_point(module_scope)
